# Author: Epihaius
# Date: 2019-06-21
#
# This module contains a class to generate torus primitives.

from .base import *


class TorusMaker(ModelMaker):
    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, pos):
        self._center = pos

    @property
    def ring_radius(self):
        return self._ring_radius

    @ring_radius.setter
    def ring_radius(self, radius):
        self._ring_radius = radius

    @property
    def section_radius(self):
        return self._section_radius

    @section_radius.setter
    def section_radius(self, radius):
        self._section_radius = radius

    @property
    def smooth_ring(self):
        return self._smooth_ring

    @smooth_ring.setter
    def smooth_ring(self, smooth):
        self._smooth_ring = smooth

    @property
    def smooth_section(self):
        return self._smooth_section

    @smooth_section.setter
    def smooth_section(self, smooth):
        self._smooth_section = smooth

    @property
    def ring_slice(self):
        return self._ring_slice

    @ring_slice.setter
    def ring_slice(self, angle):
        self._ring_slice = angle

    @property
    def section_slice(self):
        return self._section_slice

    @section_slice.setter
    def section_slice(self, angle):
        self._section_slice = angle

    @property
    def rotation(self):
        return self._rotation

    @rotation.setter
    def rotation(self, angle):
        self._rotation = angle

    @property
    def twist(self):
        return self._twist

    @twist.setter
    def twist(self, angle):
        self._twist = angle

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, thickness):
        self._thickness = thickness

    def __init__(
        self,
        center=None,
        ring_radius=2.,
        section_radius=1.,
        segments=None,
        smooth_ring=True,
        smooth_section=True,
        ring_slice=0.,
        section_slice=0.,
        rotation=0.,
        twist=0.,
        thickness=None,
        inverted=False,
        vertex_color=None,
        has_uvs=True,
        tex_units=None,
        tex_offset=None,
        tex_rotation=None,
        tex_scale=None
    ):
        """
        This class generates torus model primitives with the given parameters:

            center (sequence or None):
                the position of the center of the geometry in object space;
                default is at the origin (the object-space (0., 0., 0.) point);
            ring_radius (float):
                the radius of the ring; cannot be negative;
                default = 2.;
            section_radius (float):
                the radius of the cross-sections perpendicular to the ring;
                cannot be negative;
                default = 1.;
            segments (dict of ints):
                the number of surface subdivisions:
                    "ring":
                        subdivisions along the ring;
                        minimum = 3, default = 20;
                    "section":
                        subdivisions along the cross-sections;
                        minimum = 3, default = 10;
                    "ring_slice_start_cap":
                        radial subdivisions of the cap at the start of the
                        ring slice;
                        minimum = 0 (no cap), default = 1;
                    "ring_slice_end_cap":
                        radial subdivisions of the cap at the end of the
                        ring slice;
                        minimum = 0 (no cap), default = 1;
                    "section_slice_start_cap":
                        radial subdivisions of the cap at the start of the
                        section slice;
                        minimum = 0 (no cap), default = 1;
                    "section_slice_end_cap":
                        radial subdivisions of the cap at the end of the
                        section slice;
                        minimum = 0 (no cap), default = 1;
            smooth_ring (bool):
                whether the surface of the mantle along the ring should appear
                smooth or faceted;
                default is True;
            smooth_section (bool):
                whether the surface of the mantle along the cross-sections should
                appear smooth or faceted;
                default is True;
            ring_slice (float):
                the angle of the ring pie slice removed from the torus, in degrees;
                must be in [0., 360.] range; starts at the positive X-axis;
                default = 0. (no slice);
            section_slice (float):
                the angle of the section pie slice removed from the torus, in degrees;
                must be in [0., 360.] range; starts at the positive X-axis;
                default = 0. (no slice);
            rotation (float):
                the angle at which the cross-sections are rotated about the ring,
                in degrees;
                default = 0.;
            twist (float):
                like rotation, but for the last cross-section relative to the first
                cross-section;
                default = 0.;
            thickness (float or None):
                radial offset of inner torus;
                results in a curved tube with an inner section radius equal to
                section_radius minus thickness;
                must be in [0., section_radius] range;
                default = None (no inner torus).

        The parameters common to all primitive types are documented in the
        ModelMaker base class.

        The surfaces of this primitive type can be referred to in the relevant
        parameters and properties as follows:
            "main", "ring_slice_start_cap", "ring_slice_end_cap",
            "section_slice_start_cap", "section_slice_end_cap", "inner_main".

        Ranges of vertex indices of each surface can be retrieved through the
        vertex_ranges property, as a dict of (start_index, end_index) tuples
        (empty tuple if the surface was not created), with end_index not
        included in the range.

        """

        surface_ids = (
            "main", "ring_slice_start_cap", "ring_slice_end_cap", "section_slice_start_cap", "section_slice_end_cap",
            "inner_main"
        )

        ModelMaker.__init__(
            self, segments, inverted, vertex_color, has_uvs, tex_units, tex_offset, tex_rotation, tex_scale, surface_ids
        )

        self._center = center
        self._ring_radius = ring_radius
        self._section_radius = section_radius
        self._smooth_ring = smooth_ring
        self._smooth_section = smooth_section
        self._ring_slice = ring_slice
        self._section_slice = section_slice
        self._rotation = rotation
        self._twist = twist
        self._thickness = thickness

    def reset(self):

        ModelMaker.reset(self)

        self._center = None
        self._ring_radius = 2.
        self._section_radius = 1.
        self._smooth_ring = True
        self._smooth_section = True
        self._ring_slice = 0.
        self._section_slice = 0.
        self._rotation = 0.
        self._twist = 0.
        self._thickness = None

    def __add_cap_data(
        self, segs, point, vec, points, uvs, has_uvs, u_start, mirror_v, tex_size, radius, ring_arc, section_arc, mat
    ):

        for i in range(segs + 1):

            points.append(point + vec * i / segs)

            if has_uvs:

                u = u_start
                v = .5 * vec.length() / radius * i / segs

                if mirror_v:
                    v = 1. - v

                if tex_size:
                    u *= ring_arc / tex_size[0]
                    v *= section_arc / tex_size[1]
                if mat:
                    u, v = mat.xform_point(Point2(u, v))

            else:

                u = v = 0.

            uvs.append((u, v))

    def generate(self):

        center = (0., 0., 0.) if self._center is None else self._center
        ring_radius = max(0., self._ring_radius)
        section_radius = max(.001, self._section_radius)
        segs = {} if self._segments is None else self._segments
        segs_r = max(3, segs.get("ring", 20))
        segs_s = max(3, segs.get("section", 10))
        segs_rssc = max(0, segs.get("ring_slice_start_cap", 1))
        segs_rsec = max(0, segs.get("ring_slice_end_cap", 1))
        segs_sssc = max(0, segs.get("section_slice_start_cap", 1))
        segs_ssec = max(0, segs.get("section_slice_end_cap", 1))
        smooth_ring = self._smooth_ring
        smooth_section = self._smooth_section
        smooth = smooth_ring or smooth_section
        ring_slice = max(0., min(360., self._ring_slice))
        ring_slice_radians = pi * ring_slice / 180.
        section_slice = max(0., min(360., self._section_slice))
        section_slice_radians = pi * section_slice / 180.
        rotation = self._rotation
        twist = self._twist
        rot = pi * rotation / 180.
        inverted = self._inverted
        tw = pi * twist / (180. * segs_r) * (-1 if inverted else 1.)
        twist_angle = -tw * segs_r if inverted else 0.
        delta_angle_h = pi * ((360. - ring_slice) / 180.) / segs_r
        delta_angle_v = pi * ((360. - section_slice) / 180.) / segs_s
        thickness = section_radius if self._thickness is None \
                    else max(0., min(section_radius, self._thickness))
        inner_radius = section_radius - thickness
        has_uvs = self._has_uvs
        tex_units = self._tex_units
        tex_offset = self._tex_offset
        tex_rotation = self._tex_rotation
        tex_scale = self._tex_scale
        self._vert_ranges = vert_ranges = {
            "main": (),
            "ring_slice_start_cap": (),
            "ring_slice_end_cap": (),
            "section_slice_start_cap": (),
            "section_slice_end_cap": (),
            "inner_main": ()
        }
        stride = 8 if has_uvs else 6  # number of floats on each vertex data row
        values = array.array("f", [])
        indices = array.array("I", [])
        verts = []
        index_offset = 0

        # Define the main quad vertices

        if section_slice and (segs_sssc or segs_ssec) and thickness:
            # initialize cap data, to be gathered while defining the vertices
            # of the main quads, as those along the cap borders share position
            # and uvs
            cap_data = {"start": [], "end": []}
            cap_normal = Vec3.down() if inverted else Vec3.up()
            x = delta_angle_h * ring_radius * (-1. if inverted else 1.)
            y = tw * section_radius
            angle = atan2(y, x)
            quat = Quat()
            quat.set_from_axis_angle_rad(angle, Vec3.right())
            cap_normal = quat.xform(cap_normal)

        if has_uvs:

            ring_arc = (2 * pi - ring_slice_radians) * ring_radius
            section_arc = (2 * pi - section_slice_radians) * section_radius

            if tex_units and "main" in tex_units:
                tex_size = tex_units["main"]
            else:
                tex_size = None

            if tex_units and "section_slice_start_cap" in tex_units:
                tex_size_sssc = tex_units["section_slice_start_cap"]
            else:
                tex_size_sssc = None

            if tex_units and "section_slice_end_cap" in tex_units:
                tex_size_ssec = tex_units["section_slice_end_cap"]
            else:
                tex_size_ssec = None

            mat = self._get_tex_xform("main")
            mat_sssc = self._get_tex_xform("section_slice_start_cap")
            mat_ssec = self._get_tex_xform("section_slice_end_cap")

        for i in range(segs_r + 1):

            angle_h = delta_angle_h * i + (0. if inverted else ring_slice_radians)
            c = cos(angle_h)
            s = sin(angle_h) * (-1. if inverted else 1.)

            if has_uvs:
                u = u_sc = i / segs_r
                if tex_size:
                    u *= ring_arc / tex_size[0]
                u_start = u
            else:
                u = u_sc = 0.

            if section_slice and (segs_sssc or segs_ssec) and thickness:
                quat_h = Quat()
                quat_h.set_from_axis_angle_rad(angle_h, Vec3.down() if inverted else Vec3.up())

            for j in range(segs_s + 1):

                angle_v = rot + twist_angle + delta_angle_v * j + section_slice_radians
                r = ring_radius - section_radius * cos(angle_v)
                x = r * c
                y = r * s
                z = section_radius * sin(angle_v)
                nx = x - ring_radius * c
                ny = y - ring_radius * s

                if smooth:
                    normal = Vec3(nx, ny, z).normalized() * (-1. if inverted else 1.)

                if has_uvs:
                    v = 1. - j / segs_s
                    if tex_size:
                        v *= section_arc / tex_size[1]
                    if mat:
                        u, v = mat.xform_point(Point2(u_start, v))
                else:
                    v = 0.

                vert = {"pos": (x, y, z), "normal": normal if smooth else None, "uv": (u, v)}
                verts.append(vert)

                if j in (0, segs_s) and section_slice and (segs_sssc or segs_ssec) and thickness:

                    quat_v = Quat()
                    quat_v.set_from_axis_angle_rad(angle_v, Vec3.forward())
                    n_vec = (quat_v * quat_h).xform(cap_normal)
                    p1 = Point3(x, y, z)
                    x = ring_radius * c
                    y = ring_radius * s
                    p2 = Point3(x, y, 0.)
                    r_vec = (p2 - p1).normalized() * thickness
                    points = []
                    uvs = []

                    if j == 0:
                        self.__add_cap_data(
                            segs_sssc, p1, r_vec, points, uvs, has_uvs, u_sc, False, tex_size_sssc, section_radius,
                            ring_arc, section_arc, mat_sssc
                        )

                        n_vec *= 1. if inverted else -1.
                        cap_data["start"].append((points, uvs, n_vec))

                    elif j == segs_s:
                        self.__add_cap_data(
                            segs_ssec, p1, r_vec, points, uvs, has_uvs, u_sc, True, tex_size_ssec, section_radius,
                            ring_arc, section_arc, mat_ssec
                        )

                        n_vec *= -1. if inverted else 1.
                        cap_data["end"].append((points, uvs, n_vec))

                if not smooth_section and 0 < j < segs_s:
                    verts.append(vert.copy())

            twist_angle += tw

            if not smooth_ring and 0 < i < segs_r:
                # duplicate latest added cross-section vertices
                n = segs_s + 1 if smooth_section else segs_s * 2
                verts.extend(v.copy() for v in verts[-n:])

        # Define the vertex order of the main quads

        n = segs_s + 1 if smooth_section else segs_s * 2
        f = 1 if smooth_section else 2

        for i in range(1, segs_r + 1):

            for j in range(0, segs_s * f, f):

                vi1 = i * n + j + index_offset
                vi2 = vi1 - n
                vi3 = vi2 + 1
                vi4 = vi1 + 1
                indices.extend((vi1, vi2, vi4) if inverted else (vi1, vi2, vi3))
                indices.extend((vi2, vi3, vi4) if inverted else (vi1, vi3, vi4))

                if not smooth:
                    self._make_flat_shaded((vi1, vi2, vi3, vi4), verts)
                elif not smooth_ring:
                    self._average_normals(vi1, vi2, verts)
                    self._average_normals(vi3, vi4, verts)
                elif not smooth_section:
                    self._average_normals(vi1, vi4, verts)
                    self._average_normals(vi2, vi3, verts)

            if not smooth_ring:
                index_offset += segs_s + 1 if smooth_section else segs_s * 2

        vert_ranges["main"] = (0, len(verts))

        if ring_slice and thickness:

            # Define the ring slice cap vertices

            for cap_id in ("start", "end"):

                segs_sc = segs_rssc if cap_id == "start" else segs_rsec

                if segs_sc:

                    vertex_count = len(verts)
                    twist_angle = tw * segs_r if cap_id == "start" else 0.
                    twist_angle *= -1. if inverted else 1.

                    if cap_id == "start":
                        normal = (0., -1. if inverted else 1., 0.)
                    else:
                        angle_h = delta_angle_h * segs_r
                        c_h = cos(angle_h)
                        s_h = -sin(angle_h)
                        normal = Vec3(s_h, -c_h, 0.) * (-1. if inverted else 1.)

                    if has_uvs:
                        cap_name = "ring_slice_{}_cap".format(cap_id)
                        if tex_units and cap_name in tex_units:
                            tex_size = tex_units[cap_name]
                        else:
                            tex_size = None
                        mat = self._get_tex_xform(cap_name)

                    if not inner_radius:

                        # Define the ring slice cap triangle vertices

                        index_offset = vertex_count

                        if cap_id == "start":
                            pos = (ring_radius, 0., 0.)
                        else:
                            pos = (ring_radius * c_h, ring_radius * s_h, 0.)

                        u = v = .5

                        if has_uvs and mat:
                            u, v = mat.xform_point(Point2(u, v))

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                        radius = section_radius / segs_sc

                        for i in range(segs_s + 1):

                            angle_v = rot + twist_angle + delta_angle_v * i + section_slice_radians
                            c = cos(angle_v)
                            s = sin(angle_v)
                            r_ = radius * c
                            r = ring_radius - r_
                            z = radius * s

                            if cap_id == "start":
                                pos = (r, 0., z)
                            else:
                                pos = (r * c_h, r * s_h, z)

                            if has_uvs:

                                if cap_id == "start":
                                    u = .5 + .5 * r_ / section_radius * (-1. if inverted else 1.)
                                else:
                                    u = .5 - .5 * r_ / section_radius * (-1. if inverted else 1.)

                                v = .5 + .5 * s / segs_sc

                                if tex_size:
                                    u = (u - .5) * 2. * section_radius / tex_size[0] + .5
                                    v = (v - .5) * 2. * section_radius / tex_size[1] + .5

                                if mat:
                                    u, v = mat.xform_point(Point2(u, v))

                            else:

                                u = v = 0.

                            vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                            verts.append(vert)

                        # Define the vertex order of the ring slice cap triangles

                        for i in range(index_offset + 1, index_offset + 1 + segs_s):
                            if cap_id == "start":
                                if inverted:
                                    indices.extend((index_offset, i + 1, i))
                                else:
                                    indices.extend((index_offset, i, i + 1))
                            else:
                                if inverted:
                                    indices.extend((index_offset, i, i + 1))
                                else:
                                    indices.extend((index_offset, i + 1, i))

                    # Define the vertices of the ring slice cap quads

                    n = 0 if inner_radius else 1

                    for i in range(n, segs_sc + 1 - n):

                        radius = inner_radius + thickness * (i + n) / segs_sc

                        for j in range(segs_s + 1):

                            angle_v = rot + twist_angle + delta_angle_v * j + section_slice_radians
                            c = cos(angle_v)
                            s = sin(angle_v)
                            r_ = radius * c
                            r = ring_radius - r_
                            z = radius * s

                            if cap_id == "start":
                                pos = (r, 0., z)
                            else:
                                pos = (r * c_h, r * s_h, z)

                            if has_uvs:

                                if cap_id == "start":
                                    u = .5 + .5 * r_ / section_radius * (-1. if inverted else 1.)
                                else:
                                    u = .5 - .5 * r_ / section_radius * (-1. if inverted else 1.)

                                v = .5 + .5 * z / section_radius

                                if tex_size:
                                    u = (u - .5) * 2. * section_radius / tex_size[0] + .5
                                    v = (v - .5) * 2. * section_radius / tex_size[1] + .5

                                if mat:
                                    u, v = mat.xform_point(Point2(u, v))

                            else:

                                u = v = 0.

                            vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                            verts.append(vert)

                    # Define the vertex order of the ring slice cap quads

                    index_offset = vertex_count + (segs_s + 1 if inner_radius else 1)

                    for i in range(0 if inner_radius else 1, segs_sc):
                        for j in range(segs_s):
                            vi1 = index_offset + i * (segs_s + 1) + j
                            vi2 = vi1 - segs_s - 1
                            vi3 = vi2 + 1
                            vi4 = vi1 + 1
                            if inverted:
                                indices.extend((vi1, vi2, vi3) if cap_id == "start" else (vi1, vi3, vi2))
                                indices.extend((vi1, vi3, vi4) if cap_id == "start" else (vi1, vi4, vi3))
                            else:
                                indices.extend((vi1, vi3, vi2) if cap_id == "start" else (vi1, vi2, vi3))
                                indices.extend((vi1, vi4, vi3) if cap_id == "start" else (vi1, vi3, vi4))

                    surface_id = "ring_slice_{}_cap".format(cap_id)
                    vert_ranges[surface_id] = (vertex_count, len(verts))

        if section_slice and thickness:

            # Define the section slice cap vertices

            for cap_id in ("start", "end"):

                segs_sc = segs_sssc if cap_id == "start" else segs_ssec

                if segs_sc:

                    vertex_count = len(verts)

                    for points, uvs, normal in cap_data[cap_id]:
                        for point, uv in zip(points, uvs):
                            vert = {"pos": point, "normal": normal, "uv": uv}
                            verts.append(vert)

                    # Define the vertex order of the section slice cap quads

                    index_offset = vertex_count

                    for i in range(segs_r):
                        for j in range(segs_sc):
                            vi1 = index_offset + j
                            vi2 = vi1 + segs_sc + 1
                            vi3 = vi2 + 1
                            vi4 = vi1 + 1
                            indices.extend((vi1, vi2, vi3) if cap_id == "start" else (vi1, vi3, vi2))
                            indices.extend((vi1, vi3, vi4) if cap_id == "start" else (vi1, vi4, vi3))
                        index_offset += segs_sc + 1

                    surface_id = "section_slice_{}_cap".format(cap_id)
                    vert_ranges[surface_id] = (vertex_count, len(verts))

        for vert in verts:

            values.extend(vert["pos"])
            values.extend(vert["normal"])

            if has_uvs:
                values.extend(vert["uv"])

        # Create the geometry structures

        if inner_radius:

            # If a thickness is given, an inner torus needs to be created to close
            # the surface of the model; its parameters are derived from those of
            # the outer torus and adjusted to make both torii fit together.

            segs = {
                "ring": segs_r,
                "section": segs_s,
                "ring_slice_start_cap": 0,
                "ring_slice_end_cap": 0,
                "section_slice_start_cap": 0,
                "section_slice_end_cap": 0
            }
            inner_tex_units = {} if tex_units else None

            if tex_units and "inner_main" in tex_units:
                inner_tex_units["main"] = tex_units["inner_main"]

            inner_tex_offset = {} if tex_offset else None

            if tex_offset and "inner_main" in tex_offset:
                inner_tex_offset["main"] = tex_offset["inner_main"]

            inner_tex_rot = {} if tex_rotation else None

            if tex_rotation and "inner_main" in tex_rotation:
                inner_tex_rot["main"] = tex_rotation["inner_main"]

            inner_tex_scale = {} if tex_scale else None

            if tex_scale and "inner_main" in tex_scale:
                inner_tex_scale["main"] = tex_scale["inner_main"]

            model_maker = TorusMaker(
                None,
                ring_radius,
                inner_radius,
                segs,
                smooth_ring,
                smooth_section,
                ring_slice,
                section_slice,
                rotation,
                twist,
                inverted=not inverted,
                has_uvs=has_uvs,
                tex_units=inner_tex_units,
                tex_offset=inner_tex_offset,
                tex_rotation=inner_tex_rot,
                tex_scale=inner_tex_scale
            )
            node = model_maker.generate()

            # Extend the geometry of the inner torus with the data of the outer torus

            geom = node.modify_geom(0)
            vertex_data = geom.modify_vertex_data()
            old_vert_count = vertex_data.get_num_rows()
            old_size = old_vert_count * stride
            vertex_data.set_num_rows(old_vert_count + len(verts))
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[old_size:] = values

            x, y, z = center

            if x or y or z:
                vertex_data.transform_vertices(Mat4.translate_mat(x, y, z))

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data = vertex_data.set_color(self._vertex_color)
                geom.set_vertex_data(vertex_data)

            tris_prim = geom.modify_primitive(0)
            old_row_count = tris_prim.get_num_vertices()
            new_row_count = old_row_count + len(indices)

            if new_row_count < 2**16:
                # make the array compatible with the default index format of the
                # GeomPrimitive (16-bit) if the number of vertices allows it...
                indices = array.array("H", indices)
            else:
                # ...or force the GeomPrimitive to accept more indices by setting
                # its index format to 32-bit
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_array = tris_prim.modify_vertices()
            tris_array.set_num_rows(new_row_count)
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[old_row_count:] = indices
            tris_prim.offset_vertices(old_vert_count, old_row_count, new_row_count)

            inner_range = model_maker.vertex_ranges["main"]

            if inner_range:
                vert_ranges["inner_main"] = inner_range

            for surface_name in ("main", "ring_slice_start_cap", "ring_slice_end_cap", "section_slice_start_cap",
                                 "section_slice_end_cap"):

                vert_range = vert_ranges[surface_name]

                if vert_range:
                    start, end = vert_range
                    start += old_vert_count
                    end += old_vert_count
                    vert_ranges[surface_name] = (start, end)

        else:

            if has_uvs:
                vertex_format = GeomVertexFormat.get_v3n3t2()
            else:
                vertex_format = GeomVertexFormat.get_v3n3()

            vertex_data = GeomVertexData("torus_data", vertex_format, Geom.UH_static)
            vertex_data.unclean_set_num_rows(len(verts))
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[:] = values

            x, y, z = center

            if x or y or z:
                vertex_data.transform_vertices(Mat4.translate_mat(x, y, z))

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data = vertex_data.set_color(self._vertex_color)

            tris_prim = GeomTriangles(Geom.UH_static)

            if len(indices) < 2**16:
                indices = array.array("H", indices)
            else:
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_prim = GeomTriangles(Geom.UH_static)
            tris_array = tris_prim.modify_vertices()
            tris_array.unclean_set_num_rows(len(indices))
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[:] = indices

            geom = Geom(vertex_data)
            geom.add_primitive(tris_prim)
            node = GeomNode("torus")
            node.add_geom(geom)

        return node



import numpy as np

from metadrive.envs import VaryingDynamicsEnv


def test_varying_dynamics_env():
    env = VaryingDynamicsEnv({"num_scenarios": 10})
    try:
        dys = []
        for seed in range(10):
            env.reset(seed=seed)
            for _ in range(10):
                env.step(env.action_space.sample())
            dy = env.vehicle.get_dynamics_parameters()
            print("Dynamics: ", dy)
            dy = np.array(list(dy.values()))
            if len(dys) > 0:
                for dyy in dys:
                    assert not np.all(dy == dyy)
            dys.append(dy)
    finally:
        env.close()


if __name__ == "__main__":
    test_varying_dynamics_env()



from metadrive.component.map.argoverse_map import ArgoverseMap
from metadrive.envs.metadrive_env import MetaDriveEnv


class ArgoverseEnv(MetaDriveEnv):
    raise DeprecationWarning("This env is deprecated")

    def _post_process_config(self, config):
        config = super(ArgoverseEnv, self)._post_process_config(config)
        config["vehicle_config"]["spawn_lane_index"] = ("11713", "4250", 0)
        config["vehicle_config"]["destination"] = "968"
        return config

    def _update_map(self, episode_data: dict = None):
        xcenter, ycenter = 2599.5505965123866, 1200.0214763629717
        if self.current_map is None:
            self.config["map_config"].update(
                {
                    "city": "PIT",
                    # "draw_map_resolution": 1024,
                    "center": ArgoverseMap.metadrive_vector([xcenter, ycenter]),
                    "radius": 100
                }
            )
            map = ArgoverseMap(self.config["map_config"])
            self.engine.map_manager.load_map(map)


if __name__ == "__main__":
    env = ArgoverseEnv(
        {
            "traffic_density": 0.,
            # "debug_physics_world": True,
            "pstats": True,
            "global_light": True,
            # "debug_static_world":True,
            "cull_scene": False,
            # "controller":"joystick",
            "manual_control": True,
            "use_render": True,
            "decision_repeat": 5,
            "rgb_clip": True,
            # "debug": False,
            "vehicle_config": {
                "enable_reverse": True,
                "side_detector": dict(num_lasers=2, distance=50),
                "lane_line_detector": dict(num_lasers=2, distance=50),
            }
        }
    )

    o, _ = env.reset()
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step([1.0, 0.])
        info = {}
        info["lane_index"] = env.vehicle.lane_index
        info["heading_diff"] = env.vehicle.heading_diff(env.vehicle.lane)
        # info["left_lane_index"] =
        # info["right_lane_index"]
        env.render(text=info)
    env.close()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
from __future__ import division

from panda3d.core import ComputeNode, Vec4, Vec3, PTAInt, PTAFloat

from metadrive.render_pipeline.rpcore.gui.sprite import Sprite
from metadrive.render_pipeline.rpcore.gui.text import Text

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.loader import RPLoader


class FPSChart(RPObject):
    """ Widget to show the FPS as a chart """
    def __init__(self, pipeline, parent):
        """ Inits the widget """
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._parent = parent
        self._node = self._parent.attach_new_node("FPSChartNode")
        self._create_components()

    def _create_components(self):
        """ Internal method to init the widgets components """

        # Create the buffer which stores the last FPS values
        self._storage_buffer = Image.create_buffer("FPSValues", 250, "R16")
        self._storage_buffer.set_clear_color(Vec4(0))
        self._storage_buffer.clear_image()

        self._store_index = PTAInt.empty_array(1)
        self._store_index[0] = 0

        self._current_ftime = PTAFloat.empty_array(1)
        self._current_ftime[0] = 16.0

        self._chart_ms_max = PTAFloat.empty_array(1)
        self._chart_ms_max[0] = 40

        # Create the texture where the gui component is rendered inside
        self._display_tex = Image.create_2d("FPSChartRender", 250, 120, "RGBA8")
        self._display_tex.set_clear_color(Vec4(0))
        self._display_tex.clear_image()
        self._display_img = Sprite(image=self._display_tex, parent=self._node, w=250, h=120, x=10, y=10)

        # Defer the further loading
        Globals.base.taskMgr.doMethodLater(0.3, self._late_init, "FPSChartInit")

    def _late_init(self, task):
        """ Gets called after the pipeline was initialized """
        self._display_txt = Text(text="40 ms", parent=self._node, x=20, y=25, size=13, color=Vec3(1), may_change=True)
        self._display_txt_bottom = Text(
            text="0 ms", parent=self._node, x=20, y=120, size=13, color=Vec3(1), may_change=True
        )

        # Create the shader which generates the visualization texture
        self._cshader_node = ComputeNode("FPSChartUpdateChart")
        self._cshader_node.add_dispatch(250 // 10, 120 // 4, 1)
        self._cshader_np = self._node.attach_new_node(self._cshader_node)

        self._cshader = RPLoader.load_shader("/$$rp/shader/fps_chart.compute.glsl")
        self._cshader_np.set_shader(self._cshader)
        self._cshader_np.set_shader_inputs(
            DestTex=self._display_tex,
            FPSValues=self._storage_buffer,
            index=self._store_index,
            maxMs=self._chart_ms_max
        )

        self._update_shader_node = ComputeNode("FPSChartUpdateValues")
        self._update_shader_node.add_dispatch(1, 1, 1)
        self._update_shader_np = self._node.attach_new_node(self._update_shader_node)
        self._ushader = RPLoader.load_shader("/$$rp/shader/fps_chart_update.compute.glsl")
        self._update_shader_np.set_shader(self._ushader)
        self._update_shader_np.set_shader_inputs(
            DestTex=self._storage_buffer, index=self._store_index, currentData=self._current_ftime
        )

        Globals.base.addTask(self._update, "UpdateFPSChart", sort=-50)

        return task.done

    def _update(self, task):
        """ Updates the widget """
        self._store_index[0] = (self._store_index[0] + 1) % 250
        self._current_ftime[0] = Globals.clock.get_dt() * 1000.0

        avg_fps = Globals.clock.get_average_frame_rate()
        if avg_fps > 122:
            self._chart_ms_max[0] = 10.0
        elif avg_fps > 62:
            self._chart_ms_max[0] = 20.0
        elif avg_fps > 32:
            self._chart_ms_max[0] = 40.0
        elif avg_fps > 15:
            self._chart_ms_max[0] = 70.0
        else:
            self._chart_ms_max[0] = 200.0

        self._display_txt.set_text(str(int(self._chart_ms_max[0])) + " ms")

        return task.cont



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# pylint: disable=line-too-long

from direct.stdpy.file import isfile, open

from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.loader import RPLoader


class DisplayShaderBuilder(object):  # pylint: disable=too-few-public-methods
    """ Utility class to generate shaders on the fly to display texture previews
    and also buffers """
    @classmethod
    def build(cls, texture, view_width, view_height):
        """ Builds a shader to display <texture> in a view port with the size
        <view_width> * <view_height> """
        view_width, view_height = int(view_width), int(view_height)

        cache_key = "/$$rptemp/$$TEXDISPLAY-X{}-Y{}-Z{}-TT{}-CT{}-VW{}-VH{}.frag.glsl".format(
            texture.get_x_size(), texture.get_y_size(), texture.get_z_size(), texture.get_texture_type(),
            texture.get_component_type(), view_width, view_height
        )

        # Only regenerate the file when there is no cache entry for it
        if not isfile(cache_key) or True:
            fragment_shader = cls._build_fragment_shader(texture, view_width, view_height)

            with open(cache_key, "w") as handle:
                handle.write(fragment_shader)

        return RPLoader.load_shader("/$$rp/shader/default_gui_shader.vert.glsl", cache_key)

    @classmethod
    def _build_fragment_shader(cls, texture, view_width, view_height):
        """ Internal method to build a fragment shader displaying the texture """

        sampling_code, sampler_type = cls._generate_sampling_code(texture, view_width, view_height)

        # Build actual shader
        built = """
            #version 400
            #extension GL_ARB_shading_language_420pack : enable
            #pragma include "render_pipeline_base.inc.glsl"
            in vec2 texcoord;
            out vec3 result;
            uniform int mipmap;
            uniform int slice;
            uniform float brightness;
            uniform bool tonemap;
            uniform """ + sampler_type + """ p3d_Texture0;
            void main() {
                int view_width = """ + str(view_width) + """;
                int view_height = """ + str(
            view_height
        ) + """;
                ivec2 display_coord = ivec2(texcoord * vec2(view_width, view_height));
                int int_index = display_coord.x + display_coord.y * view_width;
                """ + sampling_code + """
                result *= brightness;
                if (tonemap)
                    result = result / (1 + result);
            }
        """

        # Strip trailing spaces
        built = '\n'.join([i.strip() for i in built.split("\n")])

        return built

    @classmethod
    def _generate_sampling_code(cls, texture, view_width, view_height):  # noqa # pylint: disable=unused-argument,too-many-branches
        """ Generates the GLSL code to sample a texture and also returns the
        GLSL sampler type """

        texture_type = texture.get_texture_type()
        comp_type = texture.get_component_type()

        # Useful snippets
        int_coord = "ivec2 int_coord = ivec2(texcoord * textureSize(p3d_Texture0, mipmap).xy);"
        slice_count = "int slice_count = textureSize(p3d_Texture0, 0).z;"

        float_types = [Image.T_float, Image.T_unsigned_byte]
        int_types = [Image.T_int, Image.T_unsigned_short, Image.T_unsigned_int_24_8]

        result = "result = vec3(1, 0, 1);", "sampler2D"

        if comp_type not in float_types + int_types:
            RPObject.global_warn("DisplayShaderBuilder", "Unkown texture component type:", comp_type)

        # 2D Textures
        if texture_type == Image.TT_2d_texture:

            if comp_type in float_types:
                result = "result = textureLod(p3d_Texture0, texcoord, mipmap).xyz;", "sampler2D"

            elif comp_type in int_types:
                result = int_coord + "result = texelFetch(p3d_Texture0, int_coord, mipmap).xyz / 10.0;", "isampler2D"  # noqa

        # Buffer Textures
        elif texture_type == Image.TT_buffer_texture:

            def range_check(code):
                return "if (int_index < textureSize(p3d_Texture0)) {" + code + "} else { result = vec3(1.0, 0.6, 0.2);};"  # noqa

            if comp_type in float_types:
                result = range_check("result = texelFetch(p3d_Texture0, int_index).xyz;"), "samplerBuffer"  # noqa

            elif comp_type in int_types:
                result = range_check(
                    "result = texelFetch(p3d_Texture0, int_index).xyz / 10.0;"
                ), "isamplerBuffer"  # noqa

        # 3D Textures
        elif texture_type == Image.TT_3d_texture:

            if comp_type in float_types:
                result = slice_count + "result = textureLod(p3d_Texture0, vec3(texcoord, (0.5 + slice) / slice_count), mipmap).xyz;", "sampler3D"  # noqa

            elif comp_type in int_types:
                result = int_coord + "result = texelFetch(p3d_Texture0, ivec3(int_coord, slice), mipmap).xyz / 10.0;", "isampler3D"  # noqa

        # 2D Texture Array
        elif texture_type == Image.TT_2d_texture_array:

            if comp_type in float_types:
                result = "result = textureLod(p3d_Texture0, vec3(texcoord, slice), mipmap).xyz;", "sampler2DArray"  # noqa

            elif comp_type in int_types:
                result = int_coord + "result = texelFetch(p3d_Texture0, ivec3(int_coord, slice), mipmap).xyz / 10.0;", "isampler2DArray"  # noqa

        # Cubemap
        elif texture_type == Image.TT_cube_map:

            code = "vec3 sample_dir = get_cubemap_coordinate(slice, texcoord*2-1);\n"
            code += "result = textureLod(p3d_Texture0, sample_dir, mipmap).xyz;"
            result = code, "samplerCube"

        # Cubemap array
        elif texture_type == Image.TT_cube_map_array:
            code = "vec3 sample_dir = get_cubemap_coordinate(slice % 6, texcoord*2-1);\n"
            code += "result = textureLod(p3d_Texture0, vec4(sample_dir, slice / 6), mipmap).xyz;"
            result = code, "samplerCubeArray"

        else:
            print("WARNING: Unhandled texture type", texture_type, "in display shader builder")

        return result



import pygame

from metadrive import MetaDriveEnv
from metadrive.component.map.city_map import CityMap
from metadrive.engine.engine_utils import initialize_engine, close_engine
from metadrive.obs.top_down_renderer import draw_top_down_map


def _t(num_blocks):
    default_config = MetaDriveEnv.default_config()
    initialize_engine(default_config)
    try:
        map_config = default_config["map_config"]
        map_config.update(dict(type="block_num", config=num_blocks))
        map = CityMap(map_config)
        m = draw_top_down_map(map, return_surface=True)
        pygame.image.save(m, "test.png".format(num_blocks))
    finally:
        close_engine()


def test_build_city():
    _t(num_blocks=1)
    _t(num_blocks=3)
    _t(num_blocks=20)


if __name__ == '__main__':
    test_build_city()



import copy

import gymnasium as gym
import numpy as np

from metadrive.envs.marl_envs.marl_intersection import MultiAgentIntersectionEnv
from metadrive.manager.agent_manager import AgentManager
from metadrive.obs.state_obs import LidarStateObservation
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.utils import Config
from metadrive.utils.math import Vector, norm, clip


class CommunicationObservation(LidarStateObservation):
    @property
    def observation_space(self):
        shape = list(self.state_obs.observation_space.shape)
        if self.config["vehicle_config"]["lidar"]["num_lasers"] > 0 \
                and self.config["vehicle_config"]["lidar"]["distance"] > 0:
            # Number of lidar rays and distance should be positive!
            lidar_dim = self.config["vehicle_config"]["lidar"]["num_lasers"]
            lidar_dim += self.env.num_agents * (5 + (4 if self.config["lidar"]["add_others_navi"] else 0))
            shape[0] += lidar_dim
        return gym.spaces.Box(-0.0, 1.0, shape=tuple(shape), dtype=np.float32)

    def __init__(self, vehicle_config, env):
        # Restore this to default since we add all others info into obs
        assert vehicle_config["lidar"]["num_others"] == 0
        super(CommunicationObservation, self).__init__(vehicle_config=vehicle_config)
        self.env = env

        self.agent_name_index_mapping = {}
        self.agent_name_slot_mapping = {}

    def observe(self, vehicle):
        if self.env.num_RL_agents != self.env.num_agents:
            agent_name = self.env.agent_manager.object_to_agent(vehicle.name)
            if agent_name not in self.env.agent_manager.RL_agents:
                return None  # Do not compute observation for non RL agents.
        return super(CommunicationObservation, self).observe(vehicle=vehicle)

    def reset(self, env, vehicle=None):
        self.agent_name_index_mapping = {}
        self.agent_name_slot_mapping = {}

    def lidar_observe(self, vehicle):
        other_v_info = []
        if vehicle.lidar.available:
            cloud_points, detected_objects = vehicle.lidar.perceive(vehicle, )

            other_v_info = self.get_global_info(vehicle)

            other_v_info += self._add_noise_to_cloud_points(
                cloud_points,
                gaussian_noise=self.config["vehicle_config"]["lidar"]["gaussian_noise"],
                dropout_prob=self.config["vehicle_config"]["lidar"]["dropout_prob"]
            )

            self.cloud_points = cloud_points
            self.detected_objects = detected_objects
        return other_v_info

    def refresh_agent_name_index_mapping(self):
        num_agents = self.env.num_agents
        name_vehicle_mapping = self.env.agent_manager.active_agents
        name_vehicle_mapping = {k: name_vehicle_mapping[k] for k in sorted(name_vehicle_mapping.keys())}

        existing_agents = set(self.agent_name_index_mapping.keys())
        new_agents = set(name_vehicle_mapping.keys())

        if existing_agents == new_agents:
            return name_vehicle_mapping

        old_agent_names = existing_agents.difference(new_agents)

        # Some new vehicles in and probability old vehicle terminates.
        for old_agent_name in old_agent_names:
            # This guy is done

            old_index = self.agent_name_index_mapping.pop(old_agent_name)
            old_slot = self.agent_name_slot_mapping.pop(old_agent_name)

            # Search which new agent is not assigned yet, send old index to it.
            new_agent_names = new_agents.difference(existing_agents)
            for new_agent_name in new_agent_names:
                if new_agent_name not in self.agent_name_index_mapping:
                    # This new guy is not assigned index yet, just give it old index
                    self.agent_name_index_mapping[new_agent_name] = old_index
                    self.agent_name_slot_mapping[new_agent_name] = old_slot
                    new_agents.remove(new_agent_name)
                    break

        # If new agent still exist, assigned new index to it.
        new_agent_names = new_agents.difference(self.agent_name_index_mapping.keys())
        for k in new_agent_names:
            index = int(k.split("agent")[1])
            index = index % num_agents
            self.agent_name_slot_mapping[k] = index

            index = (index + 1) / (num_agents)  # scale to [1/num_agents, 1]. 0 left to no vehicle.
            self.agent_name_index_mapping[k] = index

        # assert len(new_agents.symmetric_difference(self.agent_name_index_mapping.keys())) == 0
        assert all(k in self.agent_name_index_mapping for k in name_vehicle_mapping)
        assert all(k in self.agent_name_slot_mapping.keys() for k in name_vehicle_mapping)
        self.agent_name_index_mapping = {
            k: self.agent_name_index_mapping[k]
            for k in sorted(self.agent_name_index_mapping.keys())
        }
        self.agent_name_slot_mapping = {
            k: self.agent_name_slot_mapping[k]
            for k in sorted(self.agent_name_slot_mapping.keys())
        }
        # print("SLOT:", self.agent_name_slot_mapping)
        # print("AGEN:", self.agent_name_index_mapping)

        return name_vehicle_mapping

    def _process_norm(self, vector, perceive_distance):
        vector_norm = norm(*vector)
        vector = Vector(vector)
        if vector_norm > perceive_distance:
            vector = vector / vector_norm * perceive_distance
        return vector

    def get_global_info(self, ego_vehicle):
        perceive_distance = ego_vehicle.lidar.perceive_distance
        # perceive_distance = 20  # m
        speed_scale = 20  # km/h

        name_vehicle_mapping = self.refresh_agent_name_index_mapping()
        add_others_navi = ego_vehicle.config["lidar"]["add_others_navi"]

        # surrounding_vehicles += [None] * num_others
        num_agents = self.env.num_agents

        if add_others_navi:
            res_size = 5 + 4
        else:
            res_size = 5
        res = [0.0] * res_size * num_agents

        for agent_name, vehicle in name_vehicle_mapping.items():
            # for slot_index, agent_name in self.agent_name_slot_mapping.items():
            #     vehicle = name_vehicle_mapping[agent_name]

            slot_index = self.agent_name_slot_mapping[agent_name]

            ego_position = ego_vehicle.position

            assert agent_name in self.agent_name_index_mapping
            res[slot_index * res_size] = self.agent_name_index_mapping[agent_name]

            # assert isinstance(vehicle, IDMVehicle or Base), "Now MetaDrive Doesn't support other vehicle type"

            relative_position = ego_vehicle.convert_to_local_coordinates(vehicle.position, ego_position)
            relative_position = self._process_norm(relative_position, perceive_distance)

            # It is possible that the centroid of other vehicle is too far away from ego but lidar shed on it.
            # So the distance may greater than perceive distance.
            res[slot_index * res_size + 1] = clip((relative_position[0] / perceive_distance + 1) / 2, 0.0, 1.0)
            res[slot_index * res_size + 2] = clip((relative_position[1] / perceive_distance + 1) / 2, 0.0, 1.0)

            relative_velocity = ego_vehicle.convert_to_local_coordinates(vehicle.velocity, ego_vehicle.velocity)
            relative_velocity = self._process_norm(relative_velocity, speed_scale)
            res[slot_index * res_size + 3] = clip((relative_velocity[0] / speed_scale + 1) / 2, 0.0, 1.0)
            res[slot_index * res_size + 4] = clip((relative_velocity[1] / speed_scale + 1) / 2, 0.0, 1.0)

            if add_others_navi:
                ckpt1, ckpt2 = vehicle.navigation.get_checkpoints()

                relative_ckpt1 = ego_vehicle.lidar._project_to_vehicle_system(ckpt1, ego_vehicle)
                relative_ckpt1 = self._process_norm(relative_ckpt1, perceive_distance)
                res[slot_index * res_size + 5] = clip((relative_ckpt1[0] / perceive_distance + 1) / 2, 0.0, 1.0)
                res[slot_index * res_size + 6] = clip((relative_ckpt1[1] / perceive_distance + 1) / 2, 0.0, 1.0)

                relative_ckpt2 = ego_vehicle.lidar._project_to_vehicle_system(ckpt2, ego_vehicle)
                relative_ckpt2 = self._process_norm(relative_ckpt2, perceive_distance)
                res[slot_index * res_size + 7] = clip((relative_ckpt2[0] / perceive_distance + 1) / 2, 0.0, 1.0)
                res[slot_index * res_size + 8] = clip((relative_ckpt2[1] / perceive_distance + 1) / 2, 0.0, 1.0)

        # assert len(res) == len(name_vehicle_mapping) * (5 + (4 if add_others_navi else 0))
        return res


class TinyInterRuleBasedPolicy(IDMPolicy):
    """No IDM and PID are used in this Policy!"""
    def __init__(self, control_object, random_seed, target_speed=10):
        super(TinyInterRuleBasedPolicy, self).__init__(control_object=control_object, random_seed=random_seed)
        self.target_speed = target_speed  # Set to 10km/h. Default is 30km/h.

    def act(self, *args, **kwargs):
        self.move_to_next_road()

        target_lane = self.routing_target_lane
        long, lat = target_lane.local_coordinates(self.control_object.position)

        increment = self.target_speed / 3.6 * \
                    self.engine.global_config["physics_world_step_size"] * \
                    self.engine.global_config["decision_repeat"]

        new_long = long + increment
        new_pos = target_lane.position(new_long, lat)

        new_heading = target_lane.heading_theta_at(new_long + 1)
        self.control_object.set_heading_theta(new_heading)

        # Note: the last_position is not correctly set since in vehicle.before_step
        # we will set last_position to position at that time, which is the latest position.
        # This leads to wrong reward for IDM agent.
        # You need to comment out the line in before_step to make it correct!
        self.control_object.last_position = self.control_object.position
        self.control_object.set_position(new_pos)

        return [0, 0]


class MixedIDMAgentManager(AgentManager):
    """In this manager, we can replace part of RL policy by IDM policy"""
    def __init__(self, init_observations, init_action_space, num_RL_agents, ignore_delay_done=None, target_speed=10):
        super(MixedIDMAgentManager, self).__init__(
            init_observations=init_observations, init_action_space=init_action_space
        )
        self.num_RL_agents = num_RL_agents
        self.RL_agents = set()
        self.dying_RL_agents = set()
        self.all_previous_RL_agents = set()
        self.ignore_delay_done = ignore_delay_done
        self.target_speed = target_speed

    def filter_RL_agents(self, source_dict, original_done_dict=None):

        # new_ret = {k: v for k, v in source_dict.items() if k in self.RL_agents}

        new_ret = {k: v for k, v in source_dict.items() if k in self.all_previous_RL_agents}

        # if len(new_ret) > self.num_RL_agents:
        #     assert len(self.RL_agents) - len(self.dying_RL_agents) == self.num_RL_agents
        # if len(new_ret) < self.num_RL_agents:
        #     print("Something wrong!")
        return new_ret

    def get_observation_spaces(self):
        ret = self.filter_RL_agents(super(MixedIDMAgentManager, self).get_observation_spaces())
        if len(ret) == 0:
            k, v = list(super(MixedIDMAgentManager, self).get_observation_spaces().items())[0]
            return {k: v}
        else:
            return ret

    def get_action_spaces(self):
        ret = self.filter_RL_agents(super(MixedIDMAgentManager, self).get_action_spaces())
        if len(ret) == 0:
            k, v = list(super(MixedIDMAgentManager, self).get_action_spaces().items())[0]
            return {k: v}
        else:
            return ret

    def finish(self, agent_name, ignore_delay_done=False):
        # ignore_delay_done = True
        if self.ignore_delay_done is not None:
            ignore_delay_done = self.ignore_delay_done
        if agent_name in self.RL_agents:
            self.dying_RL_agents.add(agent_name)
        super(MixedIDMAgentManager, self).finish(agent_name, ignore_delay_done)

    def _remove_vehicle(self, v):
        agent_name = self.object_to_agent(v.name)
        if agent_name in self.RL_agents:
            self.RL_agents.remove(agent_name)
        if agent_name in self.dying_RL_agents:
            self.dying_RL_agents.remove(agent_name)
        super(MixedIDMAgentManager, self)._remove_vehicle(v)

    def _get_vehicles(self, config_dict: dict):
        from metadrive.component.vehicle.vehicle_type import random_vehicle_type, vehicle_type
        ret = {}
        for agent_count, (agent_id, v_config) in enumerate(config_dict.items()):
            if self.engine.global_config["random_agent_model"]:
                v_type = random_vehicle_type(self.np_random)
            else:
                if v_config.get("vehicle_model", False):
                    v_type = vehicle_type[v_config["vehicle_model"]]
                else:
                    v_type = vehicle_type["default"]
            obj = self.spawn_object(v_type, vehicle_config=v_config)
            ret[agent_id] = obj
            if (len(self.RL_agents) - len(self.dying_RL_agents)) >= self.num_RL_agents:
                # policy = IDMPolicy(obj, self.generate_seed())
                policy_cls = TinyInterRuleBasedPolicy
                obj._use_special_color = False
                self.add_policy(obj.id, policy_cls, obj, self.generate_seed(), target_speed=self.target_speed)
            else:
                policy_cls = self._get_policy()
                self.RL_agents.add(agent_id)
                self.all_previous_RL_agents.add(agent_id)
                obj._use_special_color = True
                self.add_policy(obj.id, policy_cls, obj, self.generate_seed())
        return ret

    def reset(self):
        self.RL_agents.clear()
        self.dying_RL_agents.clear()
        self.all_previous_RL_agents.clear()
        super(MixedIDMAgentManager, self).reset()

    @property
    def allow_respawn(self):
        """In native implementation, we can only respawn new vehicle when the total vehicles in the scene
        is less then num_agents. But here, since we always have N-K IDM vehicles and K RL vehicles, so
        allow_respawn is always False. We set to True to allow spawning new vehicle.
        """
        if self._allow_respawn is False:
            return False

        if len(self.active_agents) < self.engine.global_config["num_agents"]:
            return True

        if len(self.RL_agents) - len(self.dying_RL_agents) < self.num_RL_agents:
            return True

        return False


class MultiAgentTinyInter(MultiAgentIntersectionEnv):
    @staticmethod
    def default_config() -> Config:
        tiny_config = dict(
            success_reward=10.0,

            # Default MARL MetaDrive setting:
            out_of_road_penalty=10,
            crash_vehicle_penalty=10,
            crash_object_penalty=10,

            # Default single-agent MetaDrive setting:
            # out_of_road_penalty=5.0,
            # crash_vehicle_penalty=5.0,
            # crash_object_penalty=5.0,
            num_agents=8,
            num_RL_agents=8,
            map_config=dict(
                exit_length=30,
                lane_num=1,
                lane_width=4,

                # Additional config to control the radius
                radius=None
            ),

            # Whether to remove dead vehicles immediately
            ignore_delay_done=True,

            # The target speed of IDM agents, if any
            target_speed=10,

            # Whether to use full global relative information as obs
            use_communication_obs=False,
        )
        return MultiAgentIntersectionEnv.default_config().update(tiny_config, allow_add_new_key=True)

    def _get_reset_return(self):
        org = super(MultiAgentTinyInter, self)._get_reset_return()

        # if self.num_RL_agents == self.num_agents:
        #     return org

        return self.agent_manager.filter_RL_agents(org)

    def step(self, actions):
        o, r, tm, tc, i = super(MultiAgentTinyInter, self).step(actions)

        # if self.num_RL_agents == self.num_agents:
        #     return o, r, tm, tc, i

        original_done_dict = copy.deepcopy(tm)
        d = self.agent_manager.filter_RL_agents(tm, original_done_dict=original_done_dict)
        if "__all__" in d:
            d.pop("__all__")
        # assert len(d) == self.agent_manager.num_RL_agents, d
        d["__all__"] = all(d.values())
        return (
            self.agent_manager.filter_RL_agents(o, original_done_dict=original_done_dict),
            self.agent_manager.filter_RL_agents(r, original_done_dict=original_done_dict),
            d,
            self.agent_manager.filter_RL_agents(tc, original_done_dict=original_done_dict),
            self.agent_manager.filter_RL_agents(i, original_done_dict=original_done_dict),
        )

    def _preprocess_actions(self, actions):
        if self.num_RL_agents == self.num_agents:
            return super(MultiAgentTinyInter, self)._preprocess_actions(actions)

        actions = {v_id: actions[v_id] for v_id in self.vehicles.keys() if v_id in self.agent_manager.RL_agents}
        return actions

    def __init__(self, config=None):
        super(MultiAgentTinyInter, self).__init__(config=config)
        self.num_RL_agents = self.config["num_RL_agents"]
        # if self.num_RL_agents == self.num_agents:  # Not using mixed traffic and only RL agents are running.
        #     pass
        # else:
        self.agent_manager = MixedIDMAgentManager(
            init_observations=self._get_observations(),
            init_action_space=self._get_action_space(),
            num_RL_agents=self.num_RL_agents,
            ignore_delay_done=self.config["ignore_delay_done"],
            target_speed=self.config["target_speed"]
        )

    def get_single_observation(self):
        if self.config["use_communication_obs"]:
            return CommunicationObservation(self.config, self)
        else:
            return LidarStateObservation(self.config)


if __name__ == '__main__':
    env = MultiAgentTinyInter(
        config={
            "num_agents": 8,
            "num_RL_agents": 8,

            # "map_config": {"radius": 50},
            # "ignore_delay_done": True,
            # "use_communication_obs": True,
            # "vehicle_config": {
            #     "show_line_to_dest": True,
            #         "lidar": {
            #         "num_others": 2,
            #         "add_others_navi": True
            #     }
            # },
            # "manual_control": True,
            # "use_render": True,

            # "debug_static_world": True,
        }
    )
    o, _ = env.reset()
    # env.engine.force_fps.toggle()
    print("vehicle num", len(env.engine.traffic_manager.vehicles))
    print("RL agent num", len(o))
    ep_success = 0
    ep_done = 0
    ep_reward_sum = 0.0
    ep_success_reward_sum = 0.0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step({k: [0.0, 1.0] for k in env.action_space.sample().keys()})
        # env.render("top_down", camera_position=(42.5, 0), film_size=(500, 500))
        vehicles = env.vehicles

        for k, v in tm.items():
            if v and k in info:
                ep_success += int(info[k]["arrive_dest"])
                ep_reward_sum += int(info[k]["episode_reward"])
                ep_done += 1
                if info[k]["arrive_dest"]:
                    ep_success_reward_sum += int(info[k]["episode_reward"])

        # if not te["__all__"]:
        #     assert sum(
        #         [env.engine.get_policy(v.name).__class__.__name__ == "EnvInputPolicy" for k, v in vehicles.items()]
        #     ) == env.config["num_RL_agents"]
        if any(tm.values()):
            print("Somebody dead.", tm, info)
            # print("Step {}. Policies: {}".format(i, {k: v['policy'] for k, v in info.items()}))
        if tm["__all__"]:
            # assert i >= 1000
            print("Reset. ", i, info)
            # break
            print(
                "Success Rate: {:.3f}, reward: {:.3f}, success reward: {:.3f}, failed reward: {:.3f}, total num {}".
                format(
                    ep_success / ep_done if ep_done > 0 else -1, ep_reward_sum / ep_done if ep_done > 0 else -1,
                    ep_success_reward_sum / ep_success if ep_success > 0 else -1,
                    (ep_reward_sum - ep_success_reward_sum) / (ep_done - ep_success) if
                    (ep_done - ep_success) > 0 else -1, ep_done
                )
            )
            break
            env.reset()
    env.close()



import gymnasium as gym
import numpy as np

from metadrive.component.vehicle_navigation_module.node_network_navigation import NodeNetworkNavigation
from metadrive.obs.observation_base import ObservationBase
from metadrive.utils.math import clip, norm


class StateObservation(ObservationBase):
    ego_state_obs_dim = 6
    """
    Use vehicle state info, navigation info and lidar point clouds info as input
    """
    def __init__(self, config):
        if config["vehicle_config"]["navigation_module"]:
            navi_dim = config["vehicle_config"]["navigation_module"].get_navigation_info_dim()
        else:
            navi_dim = NodeNetworkNavigation.get_navigation_info_dim()
        self.navi_dim = navi_dim
        super(StateObservation, self).__init__(config)

    @property
    def observation_space(self):
        # Navi info + Other states
        shape = self.ego_state_obs_dim + self.navi_dim + self.get_line_detector_dim()
        if self.config["random_agent_model"]:
            shape += 2
        return gym.spaces.Box(-0.0, 1.0, shape=(shape, ), dtype=np.float32)

    def observe(self, vehicle):
        """
        Ego states: [
                    [Distance to left yellow Continuous line,
                    Distance to right Side Walk], if NOT use lane_line detector else:
                    [Side_detector cloud_points]

                    Difference of heading between ego vehicle and current lane,
                    Current speed,
                    Current steering,
                    Throttle/brake of last frame,
                    Steering of last frame,
                    Yaw Rate,

                     [Lateral Position on current lane.], if use lane_line detector, else:
                     [lane_line_detector cloud points]
                    ], dim >= 9
        Navi info: [
                    Projection of distance between ego vehicle and checkpoint on ego vehicle's heading direction,
                    Projection of distance between ego vehicle and checkpoint on ego vehicle's side direction,
                    Radius of the lane whose end node is the checkpoint (0 if this lane is straight),
                    Clockwise (1) or anticlockwise (0) (0 if lane is straight),
                    Angle of the lane (0 if lane is straight)
                   ] * 2, dim = 10
        Since agent observes current lane info and next lane info, and two checkpoints exist, the dimension of navi info
        is 10.
        :param vehicle: BaseVehicle
        :return: Vehicle State + Navigation information
        """
        navi_info = vehicle.navigation.get_navi_info()
        ego_state = self.vehicle_state(vehicle)
        ret = np.concatenate([ego_state, navi_info])
        return ret.astype(np.float32)

    def vehicle_state(self, vehicle):
        """
        Wrap vehicle states to list
        """
        # update out of road
        info = []
        if self.config["random_agent_model"]:
            # The length of the target vehicle
            info.append(clip(vehicle.LENGTH / vehicle.MAX_LENGTH, 0.0, 1.0))

            # The width of the target vehicle
            info.append(clip(vehicle.WIDTH / vehicle.MAX_WIDTH, 0.0, 1.0))

        if hasattr(vehicle, "side_detector") and vehicle.side_detector.available:

            # If side detector (a Lidar scanning road borders) is turn on, then add the cloud points of side detector
            info += vehicle.side_detector.perceive(vehicle, vehicle.engine.physics_world.static_world).cloud_points

        else:

            # If the side detector is turn off, then add the distance to left and right road borders as state.
            lateral_to_left, lateral_to_right, = vehicle.dist_to_left_side, vehicle.dist_to_right_side
            if vehicle.navigation.map:
                total_width = float((vehicle.navigation.map.MAX_LANE_NUM + 1) * vehicle.navigation.map.MAX_LANE_WIDTH)
            else:
                total_width = 100
            lateral_to_left /= total_width
            lateral_to_right /= total_width
            info += [clip(lateral_to_left, 0.0, 1.0), clip(lateral_to_right, 0.0, 1.0)]

        if vehicle.navigation is None or vehicle.navigation.current_ref_lanes is None:
            # TODO LQY, consider adding observations
            info += [0] * 5
        else:
            current_reference_lane = vehicle.navigation.current_ref_lanes[-1]
            info += [

                # The angular difference between vehicle's heading and the lane heading at this location.
                vehicle.heading_diff(current_reference_lane),

                # The velocity of target vehicle
                clip((vehicle.speed_km_h + 1) / (vehicle.max_speed_km_h + 1), 0.0, 1.0),

                # Current steering
                clip((vehicle.steering / vehicle.MAX_STEERING + 1) / 2, 0.0, 1.0),

                # The normalized actions at last steps
                clip((vehicle.last_current_action[1][0] + 1) / 2, 0.0, 1.0),
                clip((vehicle.last_current_action[1][1] + 1) / 2, 0.0, 1.0)
            ]

        # Current angular acceleration (yaw rate)
        heading_dir_last = vehicle.last_heading_dir
        heading_dir_now = vehicle.heading
        cos_beta = heading_dir_now.dot(heading_dir_last) / (norm(*heading_dir_now) * norm(*heading_dir_last))
        beta_diff = np.arccos(clip(cos_beta, 0.0, 1.0))
        yaw_rate = beta_diff / 0.1
        info.append(clip(yaw_rate, 0.0, 1.0))

        if vehicle.lane_line_detector.available:

            # If lane line detector (a Lidar scanning current lane borders) is turn on,
            # then add the cloud points of lane line detector
            info += vehicle.lane_line_detector.perceive(vehicle, vehicle.engine.physics_world.static_world).cloud_points

        else:

            # If the lane line detector is turn off, then add the offset of current position
            # against the central of current lane to the state. If vehicle is centered in the lane, then the offset
            # is 0 and vice versa.
            _, lateral = vehicle.lane.local_coordinates(vehicle.position)
            max_lane_width = vehicle.navigation.map.MAX_LANE_WIDTH if vehicle.navigation.map else 10
            info.append(clip((lateral * 2 / max_lane_width + 1.0) / 2.0, 0.0, 1.0))

        return info

    def get_line_detector_dim(self):
        dim = 0
        dim += 2 if self.config["vehicle_config"]["side_detector"]["num_lasers"] == 0 else \
            self.config["vehicle_config"]["side_detector"]["num_lasers"]
        dim += 1 if self.config["vehicle_config"]["lane_line_detector"]["num_lasers"] == 0 else \
            self.config["vehicle_config"]["lane_line_detector"]["num_lasers"]
        return dim


class LidarStateObservation(ObservationBase):
    def __init__(self, config):
        self.state_obs = StateObservation(config)
        super(LidarStateObservation, self).__init__(config)
        self.cloud_points = None
        self.detected_objects = None

    @property
    def observation_space(self):
        shape = list(self.state_obs.observation_space.shape)
        if self.config["vehicle_config"]["lidar"]["num_lasers"] > 0 and self.config["vehicle_config"]["lidar"][
                "distance"] > 0:
            # Number of lidar rays and distance should be positive!
            lidar_dim = self.config["vehicle_config"]["lidar"][
                "num_lasers"] + self.config["vehicle_config"]["lidar"]["num_others"] * 4
            if self.config["vehicle_config"]["lidar"]["add_others_navi"]:
                lidar_dim += self.config["vehicle_config"]["lidar"]["num_others"] * 4
            shape[0] += lidar_dim
        return gym.spaces.Box(-0.0, 1.0, shape=tuple(shape), dtype=np.float32)

    def observe(self, vehicle):
        """
        State observation + Navi info + 4 * closest vehicle info + Lidar points ,
        Definition of State Observation and Navi information can be found in **class StateObservation**
        Other vehicles' info: [
                              Projection of distance between ego and another vehicle on ego vehicle's heading direction,
                              Projection of distance between ego and another vehicle on ego vehicle's side direction,
                              Projection of speed between ego and another vehicle on ego vehicle's heading direction,
                              Projection of speed between ego and another vehicle on ego vehicle's side direction,
                              ] * 4, dim = 16

        Lidar points: 240 lidar points surrounding vehicle, starting from the vehicle head in clockwise direction

        :param vehicle: BaseVehicle
        :return: observation in 9 + 10 + 16 + 240 dim
        """
        state = self.state_observe(vehicle)
        other_v_info = self.lidar_observe(vehicle)
        self.current_observation = np.concatenate((state, np.asarray(other_v_info)))
        ret = self.current_observation
        return ret.astype(np.float32)

    def state_observe(self, vehicle):
        return self.state_obs.observe(vehicle)

    def lidar_observe(self, vehicle):
        other_v_info = []
        if vehicle.lidar.available:
            cloud_points, detected_objects = vehicle.lidar.perceive(vehicle, )
            if self.config["vehicle_config"]["lidar"]["num_others"] > 0:
                other_v_info += vehicle.lidar.get_surrounding_vehicles_info(
                    vehicle, detected_objects, self.config["vehicle_config"]["lidar"]["num_others"],
                    self.config["vehicle_config"]["lidar"]["add_others_navi"]
                )
            other_v_info += self._add_noise_to_cloud_points(
                cloud_points,
                gaussian_noise=self.config["vehicle_config"]["lidar"]["gaussian_noise"],
                dropout_prob=self.config["vehicle_config"]["lidar"]["dropout_prob"]
            )
            self.cloud_points = cloud_points
            self.detected_objects = detected_objects
        return other_v_info

    def _add_noise_to_cloud_points(self, points, gaussian_noise, dropout_prob):
        if gaussian_noise > 0.0:
            points = np.asarray(points)
            points = np.clip(points + np.random.normal(loc=0.0, scale=gaussian_noise, size=points.shape), 0.0, 1.0)

        if dropout_prob > 0.0:
            assert dropout_prob <= 1.0
            points = np.asarray(points)
            points[np.random.uniform(0, 1, size=points.shape) < dropout_prob] = 0.0

        return list(points)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""
from __future__ import print_function
from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

from panda3d.core import Vec3

from metadrive.render_pipeline.rpcore.pynative.pointer_slot_storage import PointerSlotStorage
from metadrive.render_pipeline.rpcore.pynative.gpu_command import GPUCommand

MAX_LIGHT_COUNT = 65535
MAX_SHADOW_SOURCES = 2048


class InternalLightManager(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    def __init__(self):
        self._lights = PointerSlotStorage(MAX_LIGHT_COUNT)
        self._shadow_sources = PointerSlotStorage(MAX_SHADOW_SOURCES)
        self._cmd_list = None
        self._shadow_manager = None
        self._camera_pos = Vec3(0)
        self._shadow_update_distance = 100.0

    def get_max_light_index(self):
        return self._lights.get_max_index()

    max_light_index = property(get_max_light_index)

    def get_num_lights(self):
        return self._lights.get_num_entries()

    num_lights = property(get_num_lights)

    def get_num_shadow_sources(self):
        return self._shadow_sources.get_num_entries()

    num_shadow_sources = property(get_num_shadow_sources)

    def set_shadow_manager(self, shadow_manager):
        self._shadow_manager = shadow_manager

    def get_shadow_manager(self):
        return self._shadow_manager

    shadow_manager = property(get_shadow_manager, set_shadow_manager)

    def set_command_list(self, cmd_list):
        self._cmd_list = cmd_list

    def set_camera_pos(self, pos):
        self._camera_pos = pos

    def set_shadow_update_distance(self, dist):
        self._shadow_update_distance = dist

    def add_light(self, light):
        if light.has_slot():
            print("ERROR: Cannot add light since it already has a slot!")
            return

        slot = self._lights.find_slot()
        if slot < 0:
            print("ERROR: Could not find a free slot for a new light!")
            return

        light.assign_slot(slot)
        self._lights.reserve_slot(slot, light)

        if light.get_casts_shadows():
            self.setup_shadows(light)

        self.gpu_update_light(light)

    def setup_shadows(self, light):
        light.init_shadow_sources()
        light.update_shadow_sources()

        num_sources = light.get_num_shadow_sources()
        base_slot = self._shadow_sources.find_consecutive_slots(num_sources)
        if base_slot < 0:
            print("ERROR: Failed to find slot for shadow sources!")
            return

        for i in range(num_sources):
            source = light.get_shadow_source(i)
            source.set_needs_update(True)
            slot = base_slot + i
            self._shadow_sources.reserve_slot(slot, source)
            source.set_slot(slot)

    def remove_light(self, light):
        assert light is not None
        if not light.has_slot():
            print("ERROR: Could not detach light, light was not attached!")
            return

        self._lights.free_slot(light.get_slot())
        self.gpu_remove_light(light)
        light.remove_slot()

        if light.get_casts_shadows():

            for i in range(light.get_num_shadow_sources()):
                source = light.get_shadow_source(i)
                if source.has_slot():
                    self._shadow_sources.free_slot(source.get_slot())
                if source.has_region():
                    self._shadow_manager.get_atlas().free_region(source.get_region())
                    source.clear_region()

            self.gpu_remove_consecutive_sources(light.get_shadow_source(0), light.get_num_shadow_sources())

            light.clear_shadow_sources()

    def gpu_remove_consecutive_sources(self, first_source, num_sources):
        cmd_remove = GPUCommand(GPUCommand.CMD_remove_sources)
        cmd_remove.push_int(first_source.get_slot())
        cmd_remove.push_int(num_sources)
        self._cmd_list.add_command(cmd_remove)

    def gpu_remove_light(self, light):
        cmd_remove = GPUCommand(GPUCommand.CMD_remove_light)
        cmd_remove.push_int(light.get_slot())
        self._cmd_list.add_command(cmd_remove)

    def gpu_update_light(self, light):
        cmd_update = GPUCommand(GPUCommand.CMD_store_light)
        cmd_update.push_int(light.get_slot())
        light.write_to_command(cmd_update)
        light.set_needs_update(False)
        self._cmd_list.add_command(cmd_update)

    def gpu_update_source(self, source):
        cmd_update = GPUCommand(GPUCommand.CMD_store_source)
        cmd_update.push_int(source.get_slot())
        source.write_to_command(cmd_update)
        self._cmd_list.add_command(cmd_update)

    def update_lights(self):
        for light in self._lights.begin():
            if light.get_needs_update():
                if light.casts_shadows:
                    light.update_shadow_sources()
            self.gpu_update_light(light)

    def update_shadow_sources(self):
        sources_to_update = []

        for source in self._shadow_sources.begin():
            # if source and source.get_needs_update():
            # sources_to_update.append(source)
            if source:
                bounds = source.get_bounds()
                distance_to_camera = (self._camera_pos - bounds.get_center()) - bounds.get_radius()
                if distance_to_camera < self._shadow_update_distance:
                    sources_to_update.append(source)
                else:
                    if source.has_region():
                        self._shadow_manager.get_atlas().free_region(source.get_region())
                        source.clear_region()

        def get_source_score(source):
            dist = (source.get_bounds().get_center() - self._camera_pos).length()
            return -dist + (10**10 if source.has_region() else 0)

        sorted_sources = list(sorted(sources_to_update, key=get_source_score))

        atlas = self._shadow_manager.get_atlas()
        update_slots = min(len(sorted_sources), self._shadow_manager.get_num_update_slots_left())

        for i in range(update_slots):
            if sorted_sources[i].has_region():
                atlas.free_region(sorted_sources[i].get_region())

        for i in range(update_slots):
            source = sorted_sources[i]

            if not self._shadow_manager.add_update(source):
                print("ERROR: Shadow manager ensured update slot, but slot is taken!")
                break

            region_size = atlas.get_required_tiles(source.get_resolution())
            new_region = atlas.find_and_reserve_region(region_size, region_size)
            new_uv_region = atlas.region_to_uv(new_region)
            source.set_region(new_region, new_uv_region)
            source.set_needs_update(False)
            self.gpu_update_source(source)

    def update(self):
        self.update_lights()
        self.update_shadow_sources()



"""
Physics Node is the subclass of BulletNode (BulletRigidBBodyNode/BulletGhostNode and so on)
Since callback method in BulletPhysicsEngine returns PhysicsNode class and sometimes we need to do some custom
calculation and tell Object about these results, inheriting from these BulletNode class will help communicate between
Physics Callbacks and Object class
"""

from panda3d.bullet import BulletRigidBodyNode, BulletGhostNode


class BaseRigidBodyNode(BulletRigidBodyNode):
    def __init__(self, base_object_name, type_name=None):
        node_name = base_object_name if type_name is None else type_name
        self.type_name = type_name
        super(BaseRigidBodyNode, self).__init__(node_name)
        self.setPythonTag(node_name, self)
        self.base_object_name = base_object_name
        self._clear_python_tag = False

    def rename(self, new_name):
        self.base_object_name = new_name

    def destroy(self):
        # This sentence is extremely important!
        self.base_object_name = None
        self.clearPythonTag(self.getName())
        self._clear_python_tag = True

    def __del__(self):
        assert self._clear_python_tag, "You should call destroy() of BaseRigidBodyNode!"


class BaseGhostBodyNode(BulletGhostNode):
    """
    Ghost node will not collide with any bodies, while contact information can still be accessed
    """
    def __init__(self, base_object_name, type_name=None):
        self.type_name = type_name
        node_name = base_object_name if type_name is None else type_name
        super(BaseGhostBodyNode, self).__init__(node_name)
        self.setPythonTag(node_name, self)
        self.base_object_name = base_object_name
        self._clear_python_tag = False

    def rename(self, new_name):
        self.base_object_name = new_name

    def destroy(self):
        # This sentence is extremely important!
        self.base_object_name = None
        self.clearPythonTag(self.getName())
        self._clear_python_tag = True

    def __del__(self):
        assert self._clear_python_tag, "You should call destroy() of BaseRigidBodyNode!"



import copy
from typing import Tuple, Union, List

import math
import numpy as np

from metadrive.component.lane.abs_lane import AbstractLane
from metadrive.component.lane.circular_lane import CircularLane
from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.road_network import Road
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.constants import PGLineType, PGLineColor, DrivableAreaProperty
from metadrive.utils.math import get_vertical_vector
from metadrive.utils.pg.utils import check_lane_on_road


def create_bend_straight(
    previous_lane: "StraightLane",
    following_lane_length,
    radius: float,
    angle: float,
    clockwise: bool = True,
    width: float = AbstractLane.DEFAULT_WIDTH,
    line_types: Tuple[PGLineType, PGLineType] = None,
    forbidden: bool = False,
    speed_limit: float = 20,
    priority: int = 0
):
    bend_direction = 1 if clockwise else -1
    center = previous_lane.position(previous_lane.length, bend_direction * radius)
    p_lateral = previous_lane.direction_lateral
    x, y = p_lateral
    start_phase = np.arctan2(y, x) + (np.pi if clockwise else 0)
    bend = CircularLane(
        center, radius, start_phase, angle, clockwise, width, line_types, forbidden, speed_limit, priority
    )
    length = 2 * radius * angle / 2
    bend_end = bend.position(length, 0)
    next_lane_heading = get_vertical_vector(bend_end - center)
    # nxt_dir = next_lane_heading[0] if not clockwise else next_lane_heading[1]
    nxt_dir = next_lane_heading[0] if not clockwise else next_lane_heading[1]
    nxt_dir = np.asarray(nxt_dir)
    following_lane_end = nxt_dir * following_lane_length + bend_end
    following_lane = StraightLane(bend_end, following_lane_end, width, line_types, forbidden, speed_limit, priority)
    return bend, following_lane


def CreateRoadFrom(
    lane: Union["AbstractLane", "StraightLane", "CircularLane"],
    lane_num: int,
    road: "Road",
    roadnet_to_add_lanes: "NodeRoadNetwork",  # mostly, block network
    roadnet_to_check_cross: "NodeRoadNetwork",  # mostly, previous global_network
    toward_smaller_lane_index: bool = True,
    ignore_start: str = None,
    ignore_end: str = None,
    center_line_type=PGLineType.CONTINUOUS,  # Identical to Block.CENTER_LINE_TYPE
    detect_one_side=True,
    side_lane_line_type=PGLineType.SIDE,
    inner_lane_line_type=PGLineType.BROKEN,
    center_line_color=PGLineColor.YELLOW,
    ignore_intersection_checking=None
) -> bool:
    """
        | | | |
        | | | |
        | | | |
        | | | |
    <-----smaller direction = inside direction
    Usage: give the far left lane, then create lane_num lanes including itself
    :return if the lanes created cross other lanes
    """
    lane_num -= 1  # include lane itself
    origin_lane = lane
    lanes = []
    lane_width = lane.width_at(0)
    for i in range(lane_num, 0, -1):
        side_lane = copy.copy(lane)
        if isinstance(lane, StraightLane):
            width = -lane_width if toward_smaller_lane_index else lane_width
            start = side_lane.position(0, width)
            end = side_lane.position(side_lane.length, width)
            side_lane.start = start
            side_lane.end = end
        elif isinstance(lane, CircularLane):
            new_lane_clockwise = True if lane.is_clockwise() else False
            radius1 = lane.radius
            if not toward_smaller_lane_index:
                radius2 = radius1 - lane_width if new_lane_clockwise else radius1 + lane_width
            else:
                radius2 = radius1 + lane_width if new_lane_clockwise else radius1 - lane_width
            side_lane.radius = radius2
            side_lane.update_properties()
        if i == 1:
            side_lane.line_types = [center_line_type, inner_lane_line_type] if toward_smaller_lane_index else \
                [inner_lane_line_type, side_lane_line_type]
        else:
            side_lane.line_types = [inner_lane_line_type, inner_lane_line_type]
        lanes.append(side_lane)
        lane = side_lane
    if toward_smaller_lane_index:
        lanes.reverse()
        lanes.append(origin_lane)
        origin_lane.line_types = [inner_lane_line_type if len(lanes) > 1 else center_line_type, side_lane_line_type]
    else:
        lanes.insert(0, origin_lane)
        if len(lanes) > 1:
            line_type = origin_lane.line_types[0], lanes[-1].line_types[0]
            origin_lane.line_types = line_type
    # check the left lane and right lane
    ignore = (ignore_start, ignore_end)
    factor = (
        DrivableAreaProperty.SIDEWALK_WIDTH + DrivableAreaProperty.SIDEWALK_LINE_DIST + lane_width / 2.0
    ) * 2.0 / lane_width
    if not detect_one_side:
        # Because of side walk, the width of side walk should be consider
        no_cross = not (
            check_lane_on_road(
                roadnet_to_check_cross,
                origin_lane,
                factor,
                ignore,
                ignore_intersection_checking=ignore_intersection_checking
            ) or check_lane_on_road(
                roadnet_to_check_cross,
                lanes[0],
                -0.95,
                ignore,
                ignore_intersection_checking=ignore_intersection_checking
            )
        )
    else:
        no_cross = not check_lane_on_road(
            roadnet_to_check_cross,
            origin_lane,
            factor,
            ignore,
            ignore_intersection_checking=ignore_intersection_checking
        )
    for l in lanes:
        roadnet_to_add_lanes.add_lane(road.start_node, road.end_node, l)
    if lane_num == 0:
        lanes[-1].line_types = [center_line_type, side_lane_line_type]
    lanes[0].line_colors = [center_line_color, PGLineColor.GREY]
    return no_cross


def ExtendStraightLane(
    lane: "StraightLane", extend_length: float, line_types: (PGLineType, PGLineType)
) -> "StraightLane":
    assert isinstance(lane, StraightLane)
    new_lane = copy.copy(lane)
    start_point = lane.end
    end_point = lane.position(lane.length + extend_length, 0)
    new_lane.start = start_point
    new_lane.end = end_point
    new_lane.line_types = line_types
    new_lane.update_properties()
    return new_lane


def get_lanes_on_road(road: "Road", roadnet: "NodeRoadNetwork") -> List[AbstractLane]:
    return roadnet.graph[road.start_node][road.end_node]


def CreateAdverseRoad(
    positive_road: "Road",
    roadnet_to_get_road: "NodeRoadNetwork",  # mostly, block network
    roadnet_to_check_cross: "NodeRoadNetwork",  # mostly, previous global network
    ignore_start: str = None,
    ignore_end: str = None,
    center_line_type=PGLineType.CONTINUOUS,  # Identical to Block.CENTER_LINE_TYPE
    side_lane_line_type=PGLineType.SIDE,
    inner_lane_line_type=PGLineType.BROKEN,
    center_line_color=PGLineColor.YELLOW,
    ignore_intersection_checking=None
) -> bool:
    adverse_road = -positive_road
    lanes = get_lanes_on_road(positive_road, roadnet_to_get_road)
    reference_lane = lanes[-1]
    num = len(lanes) * 2
    width = reference_lane.width_at(0)
    if isinstance(reference_lane, StraightLane):
        start_point = reference_lane.position(lanes[-1].length, -(num - 1) * width)
        end_point = reference_lane.position(0, -(num - 1) * width)
        symmetric_lane = StraightLane(
            start_point, end_point, width, lanes[-1].line_types, reference_lane.forbidden, reference_lane.speed_limit,
            reference_lane.priority
        )
    elif isinstance(reference_lane, CircularLane):
        start_phase = reference_lane.end_phase
        angle = reference_lane.angle
        new_lane_clockwise = False if reference_lane.is_clockwise() else True
        if not new_lane_clockwise:
            radius = reference_lane.radius + (num - 1) * width
        else:
            radius = reference_lane.radius - (num - 1) * width
        symmetric_lane = CircularLane(
            reference_lane.center, radius, start_phase, angle, new_lane_clockwise, width, reference_lane.line_types,
            reference_lane.forbidden, reference_lane.speed_limit, reference_lane.priority
        )
    else:
        raise ValueError("Creating other lanes is not supported")
    success = CreateRoadFrom(
        symmetric_lane,
        int(num / 2),
        adverse_road,
        roadnet_to_get_road,
        roadnet_to_check_cross,
        ignore_start=ignore_start,
        ignore_end=ignore_end,
        side_lane_line_type=side_lane_line_type,
        inner_lane_line_type=inner_lane_line_type,
        center_line_type=center_line_type,
        center_line_color=center_line_color,
        ignore_intersection_checking=ignore_intersection_checking
    )
    positive_road.get_lanes(roadnet_to_get_road)[0].line_colors = [center_line_color, PGLineColor.GREY]
    return success


def CreateTwoWayRoad(
    road_to_change: "Road",
    roadnet_to_get_road: "NodeRoadNetwork",  # mostly, block network
    roadnet_to_check_cross: "NodeRoadNetwork",  # mostly, previous global network
    new_road_name: Road = None,
    ignore_start: str = None,
    ignore_end: str = None,
    center_line_type=PGLineType.CONTINUOUS,  # Identical to Block.CENTER_LINE_TYPE
    side_lane_line_type=PGLineType.SIDE,
    inner_lane_line_type=PGLineType.BROKEN,
    ignore_intersection_checking=None
) -> bool:
    """
    This function will add a new road in reverse direction to the road network
    Then the road will change from:
    ---------->
    ---------->
    to:
    <--------->
    <--------->
    As a result, vehicles can drive in both direction
    :return: cross or not
    """
    adverse_road = Road(road_to_change.end_node, road_to_change.start_node) if new_road_name is None else new_road_name
    lanes = get_lanes_on_road(road_to_change, roadnet_to_get_road)
    reference_lane = lanes[-1]
    num = len(lanes)
    width = reference_lane.width_at(0)
    if isinstance(reference_lane, StraightLane):
        start_point = reference_lane.position(lanes[-1].length, -(num - 1) * width)
        end_point = reference_lane.position(0, -(num - 1) * width)
        symmetric_lane = StraightLane(
            start_point, end_point, width, lanes[-1].line_types, reference_lane.forbidden, reference_lane.speed_limit,
            reference_lane.priority
        )
    elif isinstance(reference_lane, CircularLane):
        start_phase = reference_lane.end_phase
        angle = reference_lane.angle
        new_lane_clockwise = False if reference_lane.is_clockwise() else True
        if not new_lane_clockwise:
            radius = reference_lane.radius + (num - 1) * width
        else:
            radius = reference_lane.radius - (num - 1) * width
        symmetric_lane = CircularLane(
            reference_lane.center, radius, start_phase, angle, new_lane_clockwise, width, reference_lane.line_types,
            reference_lane.forbidden, reference_lane.speed_limit, reference_lane.priority
        )
    else:
        raise ValueError("Creating other lanes is not supported")
    success = CreateRoadFrom(
        symmetric_lane,
        num,
        adverse_road,
        roadnet_to_get_road,
        roadnet_to_check_cross,
        ignore_start=ignore_start,
        ignore_end=ignore_end,
        side_lane_line_type=side_lane_line_type,
        inner_lane_line_type=inner_lane_line_type,
        center_line_type=center_line_type,
        ignore_intersection_checking=ignore_intersection_checking
    )
    return success


def block_socket_merge(
    socket_1: "BlockSocket", socket_2: "BlockSocket", global_network: "NodeRoadNetwork", positive_merge: False
):
    global_network.graph[socket_1.positive_road.start_node][socket_2.negative_road.start_node] = \
        global_network.graph[socket_1.positive_road.start_node].pop(socket_1.positive_road.end_node)

    global_network.graph[socket_2.positive_road.start_node][socket_1.negative_road.start_node] = \
        global_network.graph[socket_2.positive_road.start_node].pop(socket_2.positive_road.end_node)


def create_wave_lanes(
    pre_lane, lateral_dist: float, wave_length: float, last_straight_length: float, lane_width, toward_left=True
):
    """
    Prodeuce two lanes in adverse direction
    :param pre_lane: Previous abstract lane
    :param lateral_dist: the dist moved in previous lane's lateral direction
    :param wave_length: the length of two circular lanes in the previous lane's longitude direction
    :param following_lane_length: the length of last straight lane
    :return: List[Circular lane]
    """
    angle = np.pi - 2 * np.arctan(wave_length / (2 * lateral_dist))
    radius = wave_length / (2 * math.sin(angle))
    circular_lane_1, pre_lane = create_bend_straight(
        pre_lane, 10, radius, angle, False if toward_left else True, lane_width, [PGLineType.NONE, PGLineType.NONE]
    )
    pre_lane.reset_start_end(pre_lane.position(-10, 0), pre_lane.position(pre_lane.length - 10, 0))
    circular_lane_2, straight_lane = create_bend_straight(
        pre_lane, last_straight_length, radius, angle, True if toward_left else False, lane_width,
        [PGLineType.NONE, PGLineType.NONE]
    )
    return circular_lane_1, circular_lane_2, straight_lane



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.render_stage import RenderStage

from panda3d.core import Camera, OrthographicLens, NodePath, CullFaceAttrib
from panda3d.core import DepthTestAttrib, Vec4, PTALVecBase3, Vec3, SamplerState
from panda3d.core import ColorWriteAttrib


class VoxelizationStage(RenderStage):
    """ This stage voxelizes the whole scene """

    required_inputs = ["DefaultEnvmap", "AllLightsData", "maxLightIndex"]
    required_pipes = []

    # The different states of voxelization
    S_disabled = 0
    S_voxelize_x = 1
    S_voxelize_y = 2
    S_voxelize_z = 3
    S_gen_mipmaps = 4

    def __init__(self, pipeline):
        RenderStage.__init__(self, pipeline)
        self.voxel_resolution = 256
        self.voxel_world_size = -1
        self.state = self.S_disabled
        self.create_ptas()

    def set_grid_position(self, pos):
        self.pta_next_grid_pos[0] = pos

    def create_ptas(self):
        self.pta_next_grid_pos = PTALVecBase3.empty_array(1)
        self.pta_grid_pos = PTALVecBase3.empty_array(1)

    @property
    def produced_inputs(self):
        return {"voxelGridPosition": self.pta_grid_pos}

    @property
    def produced_pipes(self):
        return {"SceneVoxels": self.voxel_grid}

    def create(self):
        # Create the voxel grid used to generate the voxels
        self.voxel_temp_grid = Image.create_3d(
            "VoxelsTemp", self.voxel_resolution, self.voxel_resolution, self.voxel_resolution, "RGBA8"
        )
        self.voxel_temp_grid.set_clear_color(Vec4(0))
        self.voxel_temp_nrm_grid = Image.create_3d(
            "VoxelsTemp", self.voxel_resolution, self.voxel_resolution, self.voxel_resolution, "R11G11B10"
        )
        self.voxel_temp_nrm_grid.set_clear_color(Vec4(0))

        # Create the voxel grid which is a copy of the temporary grid, but stable
        self.voxel_grid = Image.create_3d(
            "Voxels", self.voxel_resolution, self.voxel_resolution, self.voxel_resolution, "RGBA8"
        )
        self.voxel_grid.set_clear_color(Vec4(0))
        self.voxel_grid.set_minfilter(SamplerState.FT_linear_mipmap_linear)

        # Create the camera for voxelization
        self.voxel_cam = Camera("VoxelizeCam")
        self.voxel_cam.set_camera_mask(self._pipeline.tag_mgr.get_mask("voxelize"))
        self.voxel_cam_lens = OrthographicLens()
        self.voxel_cam_lens.set_film_size(-2.0 * self.voxel_world_size, 2.0 * self.voxel_world_size)
        self.voxel_cam_lens.set_near_far(0.0, 2.0 * self.voxel_world_size)
        self.voxel_cam.set_lens(self.voxel_cam_lens)
        self.voxel_cam_np = Globals.base.render.attach_new_node(self.voxel_cam)
        self._pipeline.tag_mgr.register_camera("voxelize", self.voxel_cam)

        # Create the voxelization target
        self.voxel_target = self.create_target("VoxelizeScene")
        self.voxel_target.size = self.voxel_resolution
        self.voxel_target.prepare_render(self.voxel_cam_np)

        # Create the target which copies the voxel grid
        self.copy_target = self.create_target("CopyVoxels")
        self.copy_target.size = self.voxel_resolution
        self.copy_target.prepare_buffer()

        # TODO! Does not work with the new render target yet - maybe add option
        # to post process region for instances?
        self.copy_target.instance_count = self.voxel_resolution
        self.copy_target.set_shader_inputs(SourceTex=self.voxel_temp_grid, DestTex=self.voxel_grid)

        # Create the target which generates the mipmaps
        self.mip_targets = []
        mip_size, mip = self.voxel_resolution, 0
        while mip_size > 1:
            mip_size, mip = mip_size // 2, mip + 1
            mip_target = self.create_target("GenMipmaps:" + str(mip))
            mip_target.size = mip_size
            mip_target.prepare_buffer()
            mip_target.instance_count = mip_size
            mip_target.set_shader_inputs(SourceTex=self.voxel_grid, sourceMip=(mip - 1))
            mip_target.set_shader_input("DestTex", self.voxel_grid, False, True, -1, mip, 0)
            self.mip_targets.append(mip_target)

        # Create the initial state used for rendering voxels
        initial_state = NodePath("VXGIInitialState")
        initial_state.set_attrib(CullFaceAttrib.make(CullFaceAttrib.M_cull_none), 100000)
        initial_state.set_attrib(DepthTestAttrib.make(DepthTestAttrib.M_none), 100000)
        initial_state.set_attrib(ColorWriteAttrib.make(ColorWriteAttrib.C_off), 100000)
        self.voxel_cam.set_initial_state(initial_state.get_state())

        Globals.base.render.set_shader_inputs(
            voxelGridPosition=self.pta_next_grid_pos, VoxelGridDest=self.voxel_temp_grid
        )

    def update(self):
        self.voxel_cam_np.show()
        self.voxel_target.active = True
        self.copy_target.active = False

        for target in self.mip_targets:
            target.active = False

        # Voxelization disable
        if self.state == self.S_disabled:
            self.voxel_cam_np.hide()
            self.voxel_target.active = False

        # Voxelization from X-Axis
        elif self.state == self.S_voxelize_x:
            # Clear voxel grid
            self.voxel_temp_grid.clear_image()
            self.voxel_cam_np.set_pos(self.pta_next_grid_pos[0] + Vec3(self.voxel_world_size, 0, 0))
            self.voxel_cam_np.look_at(self.pta_next_grid_pos[0])

        # Voxelization from Y-Axis
        elif self.state == self.S_voxelize_y:
            self.voxel_cam_np.set_pos(self.pta_next_grid_pos[0] + Vec3(0, self.voxel_world_size, 0))
            self.voxel_cam_np.look_at(self.pta_next_grid_pos[0])

        # Voxelization from Z-Axis
        elif self.state == self.S_voxelize_z:
            self.voxel_cam_np.set_pos(self.pta_next_grid_pos[0] + Vec3(0, 0, self.voxel_world_size))
            self.voxel_cam_np.look_at(self.pta_next_grid_pos[0])

        # Generate mipmaps
        elif self.state == self.S_gen_mipmaps:
            self.voxel_target.active = False
            self.copy_target.active = True
            self.voxel_cam_np.hide()

            for target in self.mip_targets:
                target.active = True

            # As soon as we generate the mipmaps, we need to update the grid position
            # as well
            self.pta_grid_pos[0] = self.pta_next_grid_pos[0]

    def reload_shaders(self):
        self.copy_target.shader = self.load_plugin_shader(
            "/$$rp/shader/default_post_process_instanced.vert.glsl", "copy_voxels.frag.glsl"
        )
        mip_shader = self.load_plugin_shader(
            "/$$rp/shader/default_post_process_instanced.vert.glsl", "generate_mipmaps.frag.glsl"
        )
        for target in self.mip_targets:
            target.shader = mip_shader

    def set_shader_input(self, *args):
        Globals.render.set_shader_input(*args)

    def set_shader_inputs(self, **kwargs):
        Globals.render.set_shader_inputs(**kwargs)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function
from panda3d.core import RenderState, ColorWriteAttrib, ShaderAttrib, BitMask32

from metadrive.render_pipeline.rplibs.six import itervalues


class TagStateManager(object):
    """ Please refer to the native C++ implementation for docstrings and comments.
    This is just the python implementation, which does not contain documentation! """
    class StateContainer(object):  # pylint: disable=too-few-public-methods
        def __init__(self, tag_name, mask, write_color):
            self.cameras = []
            self.tag_states = {}
            self.tag_name = tag_name
            self.mask = BitMask32.bit(mask)
            self.write_color = write_color

    def __init__(self, main_cam_node):
        self._main_cam_node = main_cam_node
        self._main_cam_node.node().set_camera_mask(BitMask32.bit(1))
        self.containers = {
            "shadow": self.StateContainer("Shadows", 2, False),
            "voxelize": self.StateContainer("Voxelize", 3, False),
            "envmap": self.StateContainer("Envmap", 4, True),
            "forward": self.StateContainer("Forward", 5, True),
        }

    def get_mask(self, container_name):
        if container_name == "gbuffer":
            return BitMask32.bit(1)
        return self.containers[container_name].mask

    def apply_state(self, container_name, np, shader, name, sort):
        assert shader
        state = RenderState.make_empty()
        container = self.containers[container_name]

        if not container.write_color:
            state = state.set_attrib(ColorWriteAttrib.make(ColorWriteAttrib.C_off), 10000)

        state = state.set_attrib(ShaderAttrib.make(shader, sort), sort)
        container.tag_states[name] = state
        np.set_tag(container.tag_name, name)

        for camera in container.cameras:
            camera.set_tag_state(name, state)

    def cleanup_states(self):
        self._main_cam_node.node().clear_tag_states()
        for container in itervalues(self.containers):
            for camera in container.cameras:
                camera.clear_tag_states()
            container.tag_states = {}

    def register_camera(self, container_name, source):
        container = self.containers[container_name]
        source.set_tag_state_key(container.tag_name)
        source.set_camera_mask(container.mask)
        state = RenderState.make_empty()
        if not container.write_color:
            state = state.set_attrib(ColorWriteAttrib.make(ColorWriteAttrib.C_off), 10000)
        source.set_initial_state(state)
        container.cameras.append(source)

    def unregister_camera(self, container_name, source):
        container = self.containers[container_name]
        if source not in container.cameras:
            print("Could not remove source, was never attached!")
            return

        container.cameras.remove(source)
        source.clear_tag_states()
        source.set_initial_state(RenderState.make_empty())



def encode_road_section_lane_width_id(roadId, sectionId, laneId, widthId):
    """

    Args:
      roadId:
      sectionId:
      laneId:
      widthId:

    Returns:

    """
    return ".".join([str(roadId), str(sectionId), str(laneId), str(widthId)])


class LinkIndex:
    """
    Overall index of all links in the file, save everything as successors,
    predecessors can be found via a reverse search
    """
    def __init__(self):
        self._successors = {}

    def create_from_opendrive(self, opendrive):
        """Create a LinkIndex from an OpenDrive object.

        Args:
          opendrive: OpenDrive style object.

        Returns:

        """
        self._add_junctions(opendrive)
        # Extract link information from road lanes
        for road in opendrive.roads:
            for lane_section in road.lanes.lane_sections:
                for lane in lane_section.allLanes:
                    parametric_lane_id = encode_road_section_lane_width_id(road.id, lane_section.idx, lane.id, -1)

                    # Not the last lane section? > Next lane section in same road
                    if lane_section.idx < road.lanes.getLastLaneSectionIdx():
                        successorId = encode_road_section_lane_width_id(
                            road.id, lane_section.idx + 1, lane.link.successorId, -1
                        )

                        self.add_link(parametric_lane_id, successorId, lane.id >= 0)

                    # Last lane section! > Next road in first lane section
                    # Try to get next road
                    elif (road.link.successor is not None and road.link.successor.elementType != "junction"):

                        next_road = opendrive.getRoad(road.link.successor.element_id)

                        if next_road is not None:

                            if road.link.successor.contactPoint == "start":
                                successorId = encode_road_section_lane_width_id(
                                    next_road.id, 0, lane.link.successorId, -1
                                )

                            else:  # contact point = end
                                successorId = encode_road_section_lane_width_id(
                                    next_road.id,
                                    next_road.lanes.getLastLaneSectionIdx(),
                                    lane.link.successorId,
                                    -1,
                                )
                            self.add_link(parametric_lane_id, successorId, lane.id >= 0)

                    # Not first lane section? > Previous lane section in same road
                    if lane_section.idx > 0:
                        predecessorId = encode_road_section_lane_width_id(
                            road.id, lane_section.idx - 1, lane.link.predecessorId, -1
                        )

                        self.add_link(predecessorId, parametric_lane_id, lane.id >= 0)

                    # First lane section! > Previous road
                    # Try to get previous road
                    elif (road.link.predecessor is not None and road.link.predecessor.elementType != "junction"):

                        prevRoad = opendrive.getRoad(road.link.predecessor.element_id)

                        if prevRoad is not None:

                            if road.link.predecessor.contactPoint == "start":
                                predecessorId = encode_road_section_lane_width_id(
                                    prevRoad.id, 0, lane.link.predecessorId, -1
                                )

                            else:  # contact point = end
                                predecessorId = encode_road_section_lane_width_id(
                                    prevRoad.id,
                                    prevRoad.lanes.getLastLaneSectionIdx(),
                                    lane.link.predecessorId,
                                    -1,
                                )
                            self.add_link(predecessorId, parametric_lane_id, lane.id >= 0)

    def add_link(self, parametric_lane_id, successor, reverse: bool = False):
        """

        Args:
          parametric_lane_id:
          successor:
          reverse:  (Default value = False)

        Returns:

        """

        # if reverse, call function recursively with switched parameters
        if reverse:
            self.add_link(successor, parametric_lane_id)
            return

        if parametric_lane_id not in self._successors:
            self._successors[parametric_lane_id] = []

        if successor not in self._successors[parametric_lane_id]:
            self._successors[parametric_lane_id].append(successor)

    def _add_junctions(self, opendrive):
        """

        Args:
          opendrive:

        Returns:

        """
        # add junctions to link_index
        # if contact_point is start, and laneId from connecting_road is negative
        # the connecting_road is the successor
        # if contact_point is start, and laneId from connecting_road is positive
        # the connecting_road is the predecessor
        # for contact_point == end it's exactly the other way
        for junction in opendrive.junctions:
            for connection in junction.connections:
                incoming_road = opendrive.getRoad(connection.incomingRoad)
                connecting_road = opendrive.getRoad(connection.connectingRoad)
                contact_point = connection.contactPoint

                for lane_link in connection.laneLinks:
                    if contact_point == "start":

                        # decide which lane section to use (first or last)
                        if lane_link.fromId < 0:
                            lane_section_idx = (incoming_road.lanes.getLastLaneSectionIdx())
                        else:
                            lane_section_idx = 0
                        incoming_road_id = encode_road_section_lane_width_id(
                            incoming_road.id, lane_section_idx, lane_link.fromId, -1
                        )
                        connecting_road_id = encode_road_section_lane_width_id(
                            connecting_road.id, 0, lane_link.toId, -1
                        )
                        self.add_link(incoming_road_id, connecting_road_id, lane_link.toId > 0)
                    else:
                        # decide which lane section to use (first or last)
                        if lane_link.fromId < 0:
                            lane_section_idx = 0

                        else:
                            lane_section_idx = (incoming_road.lanes.getLastLaneSectionIdx())
                        incoming_road_id = encode_road_section_lane_width_id(incoming_road.id, 0, lane_link.fromId, -1)
                        connecting_road_id = encode_road_section_lane_width_id(
                            connecting_road.id,
                            connecting_road.lanes.getLastLaneSectionIdx(),
                            lane_link.toId,
                            -1,
                        )
                        self.add_link(incoming_road_id, connecting_road_id, lane_link.toId < 0)

    def remove(self, parametric_lane_id):
        """

        Args:
          parametric_lane_id:

        """
        # Delete key
        if parametric_lane_id in self._successors:
            del self._successors[parametric_lane_id]

        # Delete all occurances in successor lists
        for successorsId, successors in self._successors.items():
            if parametric_lane_id in successors:
                self._successors[successorsId].remove(parametric_lane_id)

    def get_successors(self, parametric_lane_id: str) -> list:
        """

        Args:
          parametric_lane_id: Id of ParametricLane for which to search
            successors.

        Returns:
          List of successors belonging the the ParametricLane.
        Par
        """
        if parametric_lane_id not in self._successors:
            return []

        return self._successors[parametric_lane_id]

    def get_predecessors(self, parametric_lane_id: str) -> list:
        """

        Args:
          parametric_lane_id: Id of ParametricLane for which to search predecessors.

        Returns:
          List of predecessors of a ParametricLane.
        """
        predecessors = []
        for successor_plane_id, successors in self._successors.items():
            if parametric_lane_id not in successors:
                continue

            if successor_plane_id in predecessors:
                continue

            predecessors.append(successor_plane_id)

        return predecessors



from metadrive.envs.marl_envs import MultiAgentMetaDrive, MultiAgentBottleneckEnv, MultiAgentTollgateEnv, \
    MultiAgentParkingLotEnv, MultiAgentIntersectionEnv, MultiAgentRoundaboutEnv
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.envs.safe_metadrive_env import SafeMetaDriveEnv
from metadrive.envs.top_down_env import TopDownSingleFrameMetaDriveEnv, TopDownMetaDrive, TopDownMetaDriveEnvV2
from metadrive.envs.varying_dynamics_env import VaryingDynamicsEnv



import copy
import os

import numpy as np

from metadrive.manager.base_manager import BaseManager
from metadrive.scenario.scenario_description import ScenarioDescription as SD, MetaDriveType
from metadrive.scenario.utils import read_scenario_data, read_dataset_summary


class ScenarioDataManager(BaseManager):
    DEFAULT_DATA_BUFFER_SIZE = 100
    PRIORITY = -10

    def __init__(self):
        super(ScenarioDataManager, self).__init__()
        from metadrive.engine.engine_utils import get_engine
        engine = get_engine()

        self.store_data = engine.global_config["store_data"]
        self.directory = engine.global_config["data_directory"]
        self.num_scenarios = engine.global_config["num_scenarios"]
        self.start_scenario_index = engine.global_config["start_scenario_index"]

        # for multi-worker
        self.worker_index = self.engine.global_config["worker_index"]
        self.available_scenario_indices = [
            i for i in range(
                self.start_scenario_index + self.worker_index, self.start_scenario_index +
                self.num_scenarios, self.engine.global_config["num_workers"]
            )
        ]
        self._scenarios = {}

        # Read summary file first:
        self.summary_dict, self.summary_lookup, self.mapping = read_dataset_summary(self.directory)
        self.summary_lookup[:self.start_scenario_index] = [None] * self.start_scenario_index
        end_idx = self.start_scenario_index + self.num_scenarios
        self.summary_lookup[end_idx:] = [None] * (len(self.summary_lookup) - end_idx)

        # sort scenario for curriculum training
        self.scenario_difficulty = None
        self.sort_scenarios()

        # existence check
        assert self.start_scenario_index < len(self.summary_lookup), "Insufficient scenarios!"
        assert self.start_scenario_index + self.num_scenarios <= len(self.summary_lookup), \
            "Insufficient scenarios! Need: {} Has: {}".format(self.num_scenarios,
                                                              len(self.summary_lookup) - self.start_scenario_index)

        for p in self.summary_lookup[self.start_scenario_index:end_idx]:
            p = os.path.join(self.directory, self.mapping[p], p)
            assert os.path.exists(p), "No Data at path: {}".format(p)

        # stat
        self.coverage = [0 for _ in range(self.num_scenarios)]

    @property
    def current_scenario_summary(self):
        return self.current_scenario[SD.METADATA]

    def _get_scenario(self, i):
        assert i in self.available_scenario_indices, \
            "scenario index exceeds range, scenario index: {}, worker_index: {}".format(i, self.worker_index)
        assert i < len(self.summary_lookup)
        scenario_id = self.summary_lookup[i]
        file_path = os.path.join(self.directory, self.mapping[scenario_id], scenario_id)
        ret = read_scenario_data(file_path)
        assert isinstance(ret, SD)
        self.coverage[i - self.start_scenario_index] = 1
        return ret

    def before_reset(self):
        if not self.store_data:
            assert len(self._scenarios) <= 1, "It seems you access multiple scenarios in one episode"
            self._scenarios = {}

    def get_scenario(self, i, should_copy=False):

        _debug_memory_leak = False

        if i not in self._scenarios:

            if _debug_memory_leak:
                # inner psutil function
                def process_memory():
                    import psutil
                    import os
                    process = psutil.Process(os.getpid())
                    mem_info = process.memory_info()
                    return mem_info.rss

                cm = process_memory()

            # self._scenarios.clear_if_necessary()

            if _debug_memory_leak:
                lm = process_memory()
                print("{}:  Reset! Mem Change {:.3f}MB".format("data manager clear scenario", (lm - cm) / 1e6))
                cm = lm

            ret = self._get_scenario(i)
            self._scenarios[i] = ret

            if _debug_memory_leak:
                lm = process_memory()
                print("{}:  Reset! Mem Change {:.3f}MB".format("data manager read scenario", (lm - cm) / 1e6))
                cm = lm

        else:
            ret = self._scenarios[i]
            # print("===Don't need to get new scenario. Just return: ", i)

        if should_copy:
            return copy.deepcopy(self._scenarios[i])

        # Data Manager is the first manager that accesses  data.
        # It is proper to let it validate the metadata and change the global config if needed.

        return ret

    def get_metadata(self):
        state = super(ScenarioDataManager, self).get_metadata()
        raw_data = self.current_scenario
        state["raw_data"] = raw_data
        return state

    def transform_coordinate(self, scenario):
        raise ValueError("Deprecated now as all coordinates is right-handed now")
        if not self.engine.global_config["allow_coordinate_transform"]:
            assert scenario[SD.METADATA][SD.COORDINATE] == MetaDriveType.COORDINATE_METADRIVE, \
                "Only support MetaDrive coordinate!"
        else:
            # It supports loading WaymoData or exported data in two coordinates
            if scenario[SD.METADATA][SD.COORDINATE] == MetaDriveType.COORDINATE_WAYMO:
                self._coordinate_transform = True
            elif scenario[SD.METADATA][SD.COORDINATE] == MetaDriveType.COORDINATE_METADRIVE:
                self._coordinate_transform = False
            else:
                raise ValueError()

    @property
    def current_scenario_length(self):
        return self.current_scenario[SD.LENGTH]

    @property
    def current_scenario(self):
        return self.get_scenario(self.engine.global_random_seed)

    def sort_scenarios(self):
        """
        TODO(LQY): consider exposing this API to config
        Sort scenarios to support curriculum training. You are encouraged to customize your own sort method
        :return: sorted scenario list
        """
        if self.engine.max_level == 0:
            raise ValueError("Curriculum Level should be greater than 1")
        elif self.engine.max_level == 1:
            return

        def _score(scenario_id):
            file_path = os.path.join(self.directory, self.mapping[scenario_id], scenario_id)
            scenario = read_scenario_data(file_path)
            obj_weight = 0

            # calculate curvature
            ego_car_id = scenario[SD.METADATA][SD.SDC_ID]
            state_dict = scenario["tracks"][ego_car_id]["state"]
            valid_track = state_dict["position"][np.where(state_dict["valid"].astype(int))][..., :2]

            dir = valid_track[1:] - valid_track[:-1]
            dir = np.arctan2(dir[..., 1], dir[..., 0])
            curvature = sum(abs(dir[1:] - dir[:-1]) / np.pi) + 1

            sdc_moving_dist = SD.sdc_moving_dist(scenario)
            num_moving_objs = SD.num_moving_object(scenario, object_type=MetaDriveType.VEHICLE)
            return sdc_moving_dist * curvature + num_moving_objs * obj_weight

        start = self.start_scenario_index
        end = self.start_scenario_index + self.num_scenarios
        id_scores = [(s_id, _score(s_id)) for s_id in self.summary_lookup[start:end]]
        id_scores = sorted(id_scores, key=lambda scenario: scenario[-1])
        self.summary_lookup[start:end] = [id_score[0] for id_score in id_scores]
        self.scenario_difficulty = {id_score[0]: id_score[1] for id_score in id_scores}

    def clear_stored_scenarios(self):
        self._scenarios = {}

    @property
    def current_scenario_difficulty(self):
        return self.scenario_difficulty[self.summary_lookup[self.engine.global_random_seed]
                                        ] if self.scenario_difficulty is not None else 0

    @property
    def current_scenario_id(self):
        return self.current_scenario_summary["scenario_id"]

    @property
    def current_scenario_file_name(self):
        return self.summary_lookup[self.engine.global_random_seed]

    @property
    def data_coverage(self):
        return sum(self.coverage) / len(self.coverage) * self.engine.global_config["num_workers"]



import os

VERSION = "0.4.1.1"


def asset_version():
    asset_path = os.path.join(os.path.dirname(__file__), "assets")
    asset_version = os.path.join(asset_path, "version.txt")
    with open(asset_version, "r") as file:
        lines = file.readlines()
    return lines[0]



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.obs.top_down_obs import TopDownObservation
from metadrive.obs.top_down_obs_multi_channel import TopDownMultiChannel
from metadrive.utils import Config


class TopDownSingleFrameMetaDriveEnv(MetaDriveEnv):
    @classmethod
    def default_config(cls) -> Config:
        config = MetaDriveEnv.default_config()
        # config["vehicle_config"]["lidar"].update({"num_lasers": 0, "distance": 0})  # Remove lidar
        config.update(
            {
                "frame_skip": 5,
                "frame_stack": 3,
                "post_stack": 5,
                "rgb_clip": True,
                "resolution_size": 84,
                "distance": 30
            }
        )
        return config

    def get_single_observation(self, _=None):
        return TopDownObservation(
            self.config["vehicle_config"],
            self.config["rgb_clip"],
            onscreen=self.config["use_render"],
            max_distance=self.config["distance"]
        )


class TopDownMetaDrive(TopDownSingleFrameMetaDriveEnv):
    def get_single_observation(self, _=None):
        return TopDownMultiChannel(
            self.config["vehicle_config"],
            onscreen=self.config["use_render"],
            clip_rgb=self.config["rgb_clip"],
            frame_stack=self.config["frame_stack"],
            post_stack=self.config["post_stack"],
            frame_skip=self.config["frame_skip"],
            resolution=(self.config["resolution_size"], self.config["resolution_size"]),
            max_distance=self.config["distance"]
        )


class TopDownMetaDriveEnvV2(MetaDriveEnv):
    @classmethod
    def default_config(cls) -> Config:
        config = MetaDriveEnv.default_config()
        config["vehicle_config"]["lidar"] = {"num_lasers": 0, "distance": 0}  # Remove lidar
        config.update(
            {
                "frame_skip": 5,
                "frame_stack": 3,
                "post_stack": 5,
                "rgb_clip": True,
                "resolution_size": 84,
                "distance": 30
            }
        )
        return config

    def get_single_observation(self, _=None):
        return TopDownMultiChannel(
            self.config["vehicle_config"],
            self.config["use_render"],
            self.config["rgb_clip"],
            frame_stack=self.config["frame_stack"],
            post_stack=self.config["post_stack"],
            frame_skip=self.config["frame_skip"],
            resolution=(self.config["resolution_size"], self.config["resolution_size"]),
            max_distance=self.config["distance"]
        )


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    # Test single RGB frame
    # env = TopDownSingleFrameMetaDriveEnv(dict(use_render=True, num_scenarios=1, map="C", traffic_density=1.0))
    # env.reset()
    # for _ in range(20):
    #     o, *_ = env.step([0, 1])
    #     assert env.observation_space.contains(o)
    # for _ in range(200):
    #     o, *_ = env.step([0.01, 1])
    #     env.render()
    #     # plt.imshow(o, cmap="gray")
    #     # plt.show()
    #     # print(o.mean())
    # env.close()

    # Test multi-channel frames
    env = TopDownMetaDriveEnvV2(dict(num_scenarios=1, start_seed=5000, distance=30))
    # env = TopDownMetaDrive(dict(num_scenarios=1, map="XTO", traffic_density=0.1, frame_stack=5))
    # env = TopDownMetaDrive(dict(use_render=True, manual_control=True))
    env.reset()
    names = [
        "road_network", "navigation", "past_pos", "traffic t", "traffic t-1", "traffic t-2", "traffic t-3",
        "traffic t-4"
    ]
    for _ in range(60):
        o, *_ = env.step([-0.00, 0.2])
        assert env.observation_space.contains(o)
    for _ in range(10000):
        o, r, tm, tc, i = env.step([0.0, 1])
        print("Velocity: ", i["velocity"])

        fig, axes = plt.subplots(1, o.shape[-1], figsize=(15, 3))

        # o = env.observations[env.DEFAULT_AGENT].get_screen_window()
        # import numpy as np
        # import pygame
        # o = pygame.surfarray.array3d(o)
        # o = np.transpose(o, (1, 0, 2))
        # axes[0].imshow(o)

        for o_i in range(o.shape[-1]):
            axes[o_i].imshow(o[..., o_i], cmap="gray", vmin=0, vmax=1)
            axes[o_i].set_title(names[o_i])

        fig.suptitle("Multi-channel Top-down Observation")
        plt.show()
        print(o.mean())
    env.close()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import TransparencyAttrib, Vec3, Texture, SamplerState
from direct.gui.OnscreenImage import OnscreenImage

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.loader import RPLoader


class Sprite(RPObject):
    """ Simple wrapper arround OnscreenImage, providing a simpler interface """
    def __init__(
        self, image=None, parent=None, x=0, y=0, w=None, h=None, transparent=True, near_filter=True, any_filter=True
    ):
        """ Creates a new image, taking (x,y) as topleft coordinates.

        When near_filter is set to true, a near filter will be set to the
        texture passed. This provides sharper images.

        When any_filter is set to false, the passed image won't be modified at
        all. This enables you to display existing textures, otherwise the
        texture would get a near filter in the 3D View, too. """

        RPObject.__init__(self)

        if not isinstance(image, Texture):
            if not isinstance(image, str):
                self.warn("Invalid argument to image parameter:", image)
                return
            image = RPLoader.load_texture(image)

            if w is None or h is None:
                w, h = image.get_x_size(), image.get_y_size()
        else:
            if w is None or h is None:
                w = 10
                h = 10

        self._width, self._height = w, h
        self._initial_pos = self._translate_pos(x, y)

        self.node = OnscreenImage(
            image=image, parent=parent, pos=self._initial_pos, scale=(self._width / 2.0, 1, self._height / 2.0)
        )

        if transparent:
            self.node.set_transparency(TransparencyAttrib.M_alpha)

        tex = self.node.get_texture()

        # Apply a near filter, but only if the parent has no scale, otherwise
        # it will look weird
        if near_filter and any_filter and parent.get_sx() == 1.0:
            tex.set_minfilter(SamplerState.FT_nearest)
            tex.set_magfilter(SamplerState.FT_nearest)

        if any_filter:
            tex.set_anisotropic_degree(8)
            tex.set_wrap_u(SamplerState.WM_clamp)
            tex.set_wrap_v(SamplerState.WM_clamp)

    def get_initial_pos(self):
        """ Returns the initial position of the image. This can be used for
        animations """
        return self._initial_pos

    def pos_interval(self, *args, **kwargs):
        """ Returns a pos interval, this is a wrapper around
        NodePath.posInterval """
        return self.node.posInterval(*args, **kwargs)

    def hpr_interval(self, *args, **kwargs):
        """ Returns a hpr interval, this is a wrapper around
        NodePath.hprInterval """
        return self.node.hprInterval(*args, **kwargs)

    def color_scale_interval(self, *args, **kwargs):
        """ Returns a color scale interval, this is a wrapper around
        NodePath.colorScaleInterval """
        return self.node.colorScaleInterval(*args, **kwargs)

    def set_image(self, img):
        """ Sets the current image """
        self.node.set_image(img)

    def get_width(self):
        """ Returns the width of the image in pixels """
        return self._width

    def get_height(self):
        """ Returns the height of the image in pixels """
        return self._height

    def set_pos(self, x, y):
        """ Sets the position """
        self.node.set_pos(self._translate_pos(x, y))

    def _translate_pos(self, x, y):
        """ Converts 2d coordinates to pandas coordinate system """
        return Vec3(x + self._width / 2.0, 1, -y - self._height / 2.0)

    def set_shader(self, shader):
        """ Sets a shader to be used for rendering the image """
        self.node.set_shader(shader)

    def set_shader_input(self, *args):
        """ Sets a shader input on the image """
        self.node.set_shader_input(*args)

    def set_shader_inputs(self, **kwargs):
        """ Sets multiple shader inputs on the image """
        self.node.set_shader_inputs(**kwargs)

    def remove(self):
        """ Removes the image """
        self.node.remove()

    def hide(self):
        """ Hides the image """
        self.node.hide()

    def show(self):
        """ Shows the image if it was previously hidden """
        self.node.show()

    def is_hidden(self):
        """ Returns whether the image is hidden """
        return self.node.is_hidden()



from panda3d.core import VirtualFileSystem, Filename
from metadrive.engine.asset_loader import AssetLoader
from metadrive.component.pg_space import ParameterSpace, VehicleParameterSpace
from metadrive.component.vehicle.base_vehicle import BaseVehicle


def convert_path(pth):
    return Filename.from_os_specific(pth).get_fullpath()


class DefaultVehicle(BaseVehicle):
    PARAMETER_SPACE = ParameterSpace(VehicleParameterSpace.DEFAULT_VEHICLE)
    # LENGTH = 4.51
    # WIDTH = 1.852
    # HEIGHT = 1.19
    TIRE_RADIUS = 0.313
    TIRE_WIDTH = 0.25
    MASS = 1100
    LATERAL_TIRE_TO_CENTER = 0.815
    FRONT_WHEELBASE = 1.05234
    REAR_WHEELBASE = 1.4166
    path = ['ferra/vehicle.gltf', (1, 1, 1), (0, 0.075, 0.), (0, 0, 0)]

    @property
    def LENGTH(self):
        return 4.515  # meters

    @property
    def HEIGHT(self):
        return 1.19  # meters

    @property
    def WIDTH(self):
        return 1.852  # meters


# When using DefaultVehicle as traffic, please use this class.


class TrafficDefaultVehicle(DefaultVehicle):
    pass


class StaticDefaultVehicle(DefaultVehicle):
    PARAMETER_SPACE = ParameterSpace(VehicleParameterSpace.STATIC_DEFAULT_VEHICLE)


class XLVehicle(BaseVehicle):
    PARAMETER_SPACE = ParameterSpace(VehicleParameterSpace.XL_VEHICLE)
    # LENGTH = 5.8
    # WIDTH = 2.3
    # HEIGHT = 2.8
    TIRE_RADIUS = 0.37
    TIRE_MODEL_CORRECT = -1
    REAR_WHEELBASE = 1.075
    FRONT_WHEELBASE = 1.726
    LATERAL_TIRE_TO_CENTER = 0.931
    CHASSIS_TO_WHEEL_AXIS = 0.3
    TIRE_WIDTH = 0.5
    MASS = 1600
    LIGHT_POSITION = (-0.75, 2.7, 0.2)
    path = ['truck/vehicle.gltf', (1, 1, 1), (0, 0.25, 0.04), (0, 0, 0)]

    @property
    def LENGTH(self):
        return 5.74  # meters

    @property
    def HEIGHT(self):
        return 2.8  # meters

    @property
    def WIDTH(self):
        return 2.3  # meters


class LVehicle(BaseVehicle):
    PARAMETER_SPACE = ParameterSpace(VehicleParameterSpace.L_VEHICLE)
    # LENGTH = 4.5
    # WIDTH = 1.86
    # HEIGHT = 1.85
    TIRE_RADIUS = 0.429
    REAR_WHEELBASE = 1.218261
    FRONT_WHEELBASE = 1.5301
    LATERAL_TIRE_TO_CENTER = 0.75
    TIRE_WIDTH = 0.35
    MASS = 1300
    LIGHT_POSITION = (-0.65, 2.13, 0.3)

    path = ['lada/vehicle.gltf', (1.1, 1.1, 1.1), (0, -0.27, 0.07), (0, 0, 0)]

    @property
    def LENGTH(self):
        return 4.87  # meters

    @property
    def HEIGHT(self):
        return 1.85  # meters

    @property
    def WIDTH(self):
        return 2.046  # meters


class MVehicle(BaseVehicle):
    PARAMETER_SPACE = ParameterSpace(VehicleParameterSpace.M_VEHICLE)
    # LENGTH = 4.4
    # WIDTH = 1.85
    # HEIGHT = 1.37
    TIRE_RADIUS = 0.39
    REAR_WHEELBASE = 1.203
    FRONT_WHEELBASE = 1.285
    LATERAL_TIRE_TO_CENTER = 0.803
    TIRE_WIDTH = 0.3
    MASS = 1200
    LIGHT_POSITION = (-0.67, 1.86, 0.22)

    path = ['130/vehicle.gltf', (1, 1, 1), (0, -0.05, 0.1), (0, 0, 0)]

    @property
    def LENGTH(self):
        return 4.6  # meters

    @property
    def HEIGHT(self):
        return 1.37  # meters

    @property
    def WIDTH(self):
        return 1.85  # meters


class SVehicle(BaseVehicle):
    PARAMETER_SPACE = ParameterSpace(VehicleParameterSpace.S_VEHICLE)
    # LENGTH = 4.25
    # WIDTH = 1.7
    # HEIGHT = 1.7
    LATERAL_TIRE_TO_CENTER = 0.7
    TIRE_TWO_SIDED = True
    FRONT_WHEELBASE = 1.385
    REAR_WHEELBASE = 1.11
    TIRE_RADIUS = 0.376
    TIRE_WIDTH = 0.25
    MASS = 800
    LIGHT_POSITION = (-0.57, 1.86, 0.23)

    @property
    def path(self):
        if self.use_render_pipeline:
            # vfs = VirtualFileSystem.get_global_ptr()
            # vfs.mount(convert_path(AssetLoader.file_path("models", "beetle")), "/$$beetle_model", 0)
            return ['beetle/vehicle.bam', (0.0077, 0.0077, 0.0077), (0.04512, -0.24 - 0.04512, 1.77), (-90, -90, 0)]
        else:
            factor = 1
            return ['beetle/vehicle.gltf', (factor, factor, factor), (0, -0.2, 0.03), (0, 0, 0)]

    @property
    def LENGTH(self):
        return 4.3  # meters

    @property
    def HEIGHT(self):
        return 1.70  # meters

    @property
    def WIDTH(self):
        return 1.70  # meters


class VaryingDynamicsVehicle(DefaultVehicle):
    @property
    def WIDTH(self):
        return self.config["width"] if self.config["width"] is not None else super(VaryingDynamicsVehicle, self).WIDTH

    @property
    def LENGTH(self):
        return self.config["length"] if self.config["length"] is not None else super(
            VaryingDynamicsVehicle, self
        ).LENGTH

    @property
    def HEIGHT(self):
        return self.config["height"] if self.config["height"] is not None else super(
            VaryingDynamicsVehicle, self
        ).HEIGHT

    @property
    def MASS(self):
        return self.config["mass"] if self.config["mass"] is not None else super(VaryingDynamicsVehicle, self).MASS

    def reset(
        self,
        random_seed=None,
        vehicle_config=None,
        position=None,
        heading: float = 0.0,  # In degree!
        *args,
        **kwargs
    ):

        assert "width" not in self.PARAMETER_SPACE
        assert "height" not in self.PARAMETER_SPACE
        assert "length" not in self.PARAMETER_SPACE
        should_force_reset = False
        if vehicle_config is not None:
            if vehicle_config["width"] is not None and vehicle_config["width"] != self.WIDTH:
                should_force_reset = True
            if vehicle_config["height"] is not None and vehicle_config["height"] != self.HEIGHT:
                should_force_reset = True
            if vehicle_config["length"] is not None and vehicle_config["length"] != self.LENGTH:
                should_force_reset = True
            if "max_engine_force" in vehicle_config and \
                    vehicle_config["max_engine_force"] is not None and \
                    vehicle_config["max_engine_force"] != self.config["max_engine_force"]:
                should_force_reset = True
            if "max_brake_force" in vehicle_config and \
                    vehicle_config["max_brake_force"] is not None and \
                    vehicle_config["max_brake_force"] != self.config["max_brake_force"]:
                should_force_reset = True
            if "wheel_friction" in vehicle_config and \
                    vehicle_config["wheel_friction"] is not None and \
                    vehicle_config["wheel_friction"] != self.config["wheel_friction"]:
                should_force_reset = True
            if "max_steering" in vehicle_config and \
                    vehicle_config["max_steering"] is not None and \
                    vehicle_config["max_steering"] != self.config["max_steering"]:
                self.max_steering = vehicle_config["max_steering"]
                should_force_reset = True
            if "mass" in vehicle_config and \
                    vehicle_config["mass"] is not None and \
                    vehicle_config["mass"] != self.config["mass"]:
                should_force_reset = True

        # def process_memory():
        #     import psutil
        #     import os
        #     process = psutil.Process(os.getpid())
        #     mem_info = process.memory_info()
        #     return mem_info.rss
        #
        # cm = process_memory()

        if should_force_reset:
            self.destroy()
            self.__init__(
                vehicle_config=vehicle_config,
                name=self.name,
                random_seed=self.random_seed,
                position=position,
                heading=heading
            )

            # lm = process_memory()
            # print("{}:  Reset! Mem Change {:.3f}MB".format("1 Force Re-Init Vehicle", (lm - cm) / 1e6))
            # cm = lm

        assert self.max_steering == self.config["max_steering"]

        ret = super(VaryingDynamicsVehicle, self).reset(
            random_seed=random_seed, vehicle_config=vehicle_config, position=position, heading=heading, *args, **kwargs
        )

        # lm = process_memory()
        # print("{}:  Reset! Mem Change {:.3f}MB".format("2 Force Reset Vehicle", (lm - cm) / 1e6))
        # cm = lm

        return ret


def random_vehicle_type(np_random, p=None):
    prob = [1 / len(vehicle_type) for _ in range(len(vehicle_type))] if p is None else p
    return vehicle_type[np_random.choice(list(vehicle_type.keys()), p=prob)]


vehicle_type = {
    "s": SVehicle,
    "m": MVehicle,
    "l": LVehicle,
    "xl": XLVehicle,
    "default": DefaultVehicle,
    "static_default": StaticDefaultVehicle,
    "varying_dynamics": VaryingDynamicsVehicle
}

VaryingShapeVehicle = VaryingDynamicsVehicle

type_count = [0 for i in range(3)]


def reset_vehicle_type_count(np_random=None):
    global type_count
    if np_random is None:
        type_count = [0 for i in range(3)]
    else:
        type_count = [np_random.randint(100) for i in range(3)]


def get_vehicle_type(length, np_random=None, need_default_vehicle=False):
    if np_random is not None:
        if length <= 4:
            return SVehicle
        elif length <= 5.5:
            return [LVehicle, SVehicle, MVehicle][np_random.randint(3)]
        else:
            return [LVehicle, XLVehicle][np_random.randint(2)]
    else:
        global type_count
        # evenly sample
        if length <= 4:
            return SVehicle
        elif length <= 5.5:
            type_count[1] += 1
            vs = [LVehicle, MVehicle, SVehicle]
            if need_default_vehicle:
                vs.append(TrafficDefaultVehicle)
            return vs[type_count[1] % len(vs)]
        else:
            type_count[2] += 1
            vs = [LVehicle, XLVehicle]
            return vs[type_count[2] % len(vs)]



import numpy as np

from metadrive.constants import TerminationState
from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.obs.state_obs import LidarStateObservation

info_keys = [
    "cost", "velocity", "steering", "acceleration", "step_reward", TerminationState.CRASH_VEHICLE,
    TerminationState.OUT_OF_ROAD, TerminationState.SUCCESS
]


def _act(env, action):
    assert env.action_space.contains(action)
    obs, reward, terminated, truncated, info = env.step(action)
    assert env.observation_space.contains(obs)
    assert np.isscalar(reward)
    assert isinstance(info, dict)
    for k in info_keys:
        assert k in info


def test_obs_noise():
    env = MetaDriveEnv({"vehicle_config": {"lidar": {"gaussian_noise": 1.0, "dropout_prob": 1.0}}})
    try:
        obs, _ = env.reset()
        obs_cls = env.observations[env.DEFAULT_AGENT]
        assert isinstance(obs_cls, LidarStateObservation)
        ret = obs_cls._add_noise_to_cloud_points([0.5, 0.5, 0.5], gaussian_noise=1.0, dropout_prob=1.0)
        np.testing.assert_almost_equal(np.array(ret), 0.0)
        assert env.observation_space.contains(obs)
        _act(env, env.action_space.sample())
        for x in [-1, 0, 1]:
            env.reset()
            for y in [-1, 0, 1]:
                _act(env, [x, y])
    finally:
        env.close()
    env = MetaDriveEnv({"vehicle_config": {"lidar": {"gaussian_noise": 0.0, "dropout_prob": 0.0}}})
    try:
        obs, _ = env.reset()
        obs_cls = env.observations[env.DEFAULT_AGENT]
        assert isinstance(obs_cls, LidarStateObservation)
        ret = obs_cls._add_noise_to_cloud_points([0.5, 0.5, 0.5], gaussian_noise=0.0, dropout_prob=0.0)
        assert not np.all(np.array(ret) == 0.0)
        np.testing.assert_equal(np.array(ret), np.array([0.5, 0.5, 0.5]))
        assert env.observation_space.contains(obs)
        _act(env, env.action_space.sample())
        for x in [-1, 0, 1]:
            env.reset()
            for y in [-1, 0, 1]:
                _act(env, [x, y])
    finally:
        env.close()


if __name__ == '__main__':
    test_obs_noise()



from metadrive.envs.scenario_env import ScenarioEnv
from metadrive.engine.asset_loader import AssetLoader
import time

WAYMO_ENV_CONFIG = dict(
    # ===== Map Config =====
    data_directory=AssetLoader.file_path("waymo", return_raw_style=False),
    waymo_data_directory=None,  # for compatibility
)


class WaymoEnv(ScenarioEnv):
    @classmethod
    def default_config(cls):
        config = super(WaymoEnv, cls).default_config()
        config.update(WAYMO_ENV_CONFIG)
        return config

    def __init__(self, config=None):
        super(WaymoEnv, self).__init__(config)

    def _post_process_config(self, config):
        config = super(WaymoEnv, self)._post_process_config(config)
        if config["waymo_data_directory"] is not None:
            # overwrite
            config["data_directory"] = config["waymo_data_directory"]
        return config


if __name__ == "__main__":
    from metadrive.policy.replay_policy import ReplayEgoCarPolicy

    env = WaymoEnv(
        {
            "use_render": True,
            # "agent_policy": ReplayEgoCarPolicy,
            "manual_control": True,
            "no_traffic": False,
            # "debug":True,
            # "debug_static_world": True,
            # "no_traffic":True,
            "start_scenario_index": 0,
            "show_coordinates": True,
            # "start_scenario_index": 1000,
            # "show_coordinates": True,
            "num_scenarios": 3,
            # "data_directory": "/home/shady/Downloads/test_processed",
            "show_policy_mark": True,
            "no_static_vehicles": True,
            "reactive_traffic": True,
            "horizon": 1000,
            "vehicle_config": dict(
                # no_wheel_friction=True,
                lidar=dict(num_lasers=120, distance=50, num_others=4),
                lane_line_detector=dict(num_lasers=12, distance=50),
                side_detector=dict(num_lasers=160, distance=50)
            ),
        }
    )
    success = []
    while True:
        for i in range(3):
            env.reset(seed=i)
            while True:
                step_start = time.time()
                o, r, tm, tc, info = env.step([0, 0])
                assert env.observation_space.contains(o)
                # c_lane = env.vehicle.lane
                # long, lat, = c_lane.local_coordinates(env.vehicle.position)
                print("Step: {}, Time: {}".format(env.episode_step, time.time() - step_start))
                # if env.config["use_render"]:
                env.render(
                    text={
                        "seed": env.engine.global_seed + env.config["start_scenario_index"],
                    }
                    # mode="topdown"
                )

                if tm or tc:
                    if info["arrive_dest"]:
                        print("seed:{}, success".format(env.engine.global_random_seed))
                    break
